#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This file contains all the classes needed to perform layout.
This includes an Alignment Enum type, and the base implementation of LayoutElement
"""
import typing
from decimal import Decimal
from enum import Enum

from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.line_art.blob_factory import BlobFactory


class Alignment(Enum):
    """
    In typesetting and page layout, alignment or range is the setting of text flow or image placement relative to a page,
    column (measure), table cell, or tab.
    The type alignment setting is sometimes referred to as text alignment,
    text justification, or type justification.
    The edge of a page or column is known as a margin, and a gap between columns is known as a gutter.
    """

    LEFT = 2
    CENTERED = 3
    RIGHT = 5
    JUSTIFIED = 7

    TOP = 11
    MIDDLE = 13
    BOTTOM = 17


class LayoutElement:
    """
    This class contains the common base methods for any object that can be laid out on a Page.
    e.g. the placement of borders, margins, padding, background color, etc
    """

    def __init__(
        self,
        background_color: typing.Optional[Color] = None,
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        font_size: typing.Optional[Decimal] = None,
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: typing.Optional[Decimal] = Decimal(0),
        margin_left: typing.Optional[Decimal] = Decimal(0),
        margin_right: typing.Optional[Decimal] = Decimal(0),
        margin_top: typing.Optional[Decimal] = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        parent: typing.Optional["LayoutElement"] = None,
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        # background color
        self._background_color = background_color

        # borders
        self._border_top = border_top
        self._border_right = border_right
        self._border_bottom = border_bottom
        self._border_left = border_left

        # border radii
        assert (
            border_radius_top_right >= 0
        ), "border_radius_top_right must be a non-negative integer"
        assert (
            border_radius_top_left >= 0
        ), "border_radius_top_left must be a non-negative integer"
        assert (
            border_radius_bottom_left >= 0
        ), "border_radius_bottom_left must be a non-negative integer"
        assert (
            border_radius_bottom_right >= 0
        ), "border_radius_bottom_right must be a non-negative integer"
        self._border_radius_top_left: Decimal = border_radius_top_left
        self._border_radius_top_right: Decimal = border_radius_top_right
        self._border_radius_bottom_right: Decimal = border_radius_bottom_right
        self._border_radius_bottom_left: Decimal = border_radius_bottom_left

        # border width and color
        assert border_width >= 0
        self._border_width = border_width
        self._border_color = border_color

        # font_size
        self._font_size = font_size

        # margin
        assert margin_top is None or margin_top >= 0
        assert margin_right is None or margin_right >= 0
        assert margin_bottom is None or margin_bottom >= 0
        assert margin_left is None or margin_left >= 0
        self._margin_top = margin_top
        self._margin_right = margin_right
        self._margin_bottom = margin_bottom
        self._margin_left = margin_left

        # padding
        assert padding_top >= 0
        assert padding_right >= 0
        assert padding_bottom >= 0
        assert padding_left >= 0
        self._padding_top = padding_top
        self._padding_right = padding_right
        self._padding_bottom = padding_bottom
        self._padding_left = padding_left

        # alignment
        assert horizontal_alignment in [
            Alignment.LEFT,
            Alignment.CENTERED,
            Alignment.RIGHT,
            Alignment.JUSTIFIED,
        ]
        assert vertical_alignment in [Alignment.TOP, Alignment.MIDDLE, Alignment.BOTTOM]
        self._horizontal_alignment = horizontal_alignment
        self._vertical_alignment = vertical_alignment

        # cache
        self._previous_layout_box: typing.Optional[Rectangle] = None
        self._previous_paint_box: typing.Optional[Rectangle] = None

        # linkage (for lists, tables, etc)
        self._parent = parent

    def get_font_size(self) -> Decimal:
        """
        This function returns the font size of this LayoutElement
        """
        return self._font_size or Decimal(0)

    def get_margin_top(self) -> Decimal:
        """
        This function returns the top margin of this LayoutElement
        """
        return self._margin_top or Decimal(0)

    def get_margin_right(self) -> Decimal:
        """
        This function returns the right margin of this LayoutElement
        """
        return self._margin_right or Decimal(0)

    def get_margin_bottom(self) -> Decimal:
        """
        This function returns the bottom margin of this LayoutElement
        """
        return self._margin_bottom or Decimal(0)

    def get_margin_left(self) -> Decimal:
        """
        This function returns the left margin of this LayoutElement
        """
        return self._margin_left or Decimal(0)

    #
    # NEW RENDERING LOGIC
    #

    def get_previous_layout_box(self) -> typing.Optional[Rectangle]:
        """
        This function returns the previous result of layout of this LayoutElement
        :return:    the Rectangle that was the result of the previous layout operation
        """
        return self._previous_layout_box

    def get_previous_paint_box(self) -> typing.Optional[Rectangle]:
        """
        This function returns the previous result of painting this LayoutElement
        :return:    the Rectangle that was the result of the previous paint operation
        """
        return self._previous_paint_box

    def get_layout_box(self, available_space: Rectangle):
        """
        This function returns the previous result of layout
        :return:    the Rectangle that was the result of the previous layout operation
        """
        horizontal_border_width: Decimal = Decimal(0)
        if self._border_left:
            horizontal_border_width += self._border_width
        if self._border_right:
            horizontal_border_width += self._border_width

        vertical_border_width: Decimal = Decimal(0)
        if self._border_top:
            vertical_border_width += self._border_width
        if self._border_bottom:
            vertical_border_width += self._border_width

        cbox_available_space: Rectangle = Rectangle(
            available_space.get_x()
            + self._padding_left
            + (self._border_width if self._border_left else Decimal(0)),
            available_space.get_y()
            + self._padding_bottom
            + (self._border_width if self._border_bottom else Decimal(0)),
            max(
                Decimal(0),
                available_space.get_width()
                - self._padding_left
                - self._padding_right
                - horizontal_border_width,
            ),
            max(
                Decimal(0),
                available_space.get_height()
                - self._padding_top
                - self._padding_bottom
                - vertical_border_width,
            ),
        )

        # determine content_box
        cbox: Rectangle = self._get_content_box(cbox_available_space)

        # take into account vertical_alignment
        delta_x: Decimal = Decimal(0)
        delta_y: Decimal = Decimal(0)
        if self._vertical_alignment == Alignment.MIDDLE:
            delta_y = (cbox_available_space.get_height() - cbox.get_height()) / Decimal(
                2
            )
            cbox.y -= delta_y
        if self._vertical_alignment == Alignment.BOTTOM:
            delta_y = cbox_available_space.get_height() - cbox.get_height()
            cbox.y -= delta_y

        # take into account horizontal_alignment
        if self._horizontal_alignment == Alignment.CENTERED:
            delta_x = (cbox_available_space.get_width() - cbox.get_width()) / Decimal(2)
            cbox.x += delta_x
        if self._horizontal_alignment == Alignment.RIGHT:
            delta_x = cbox_available_space.get_width() - cbox.get_width()
            cbox.x += delta_x

        # return
        # fmt: off
        self._previous_layout_box = Rectangle(
            cbox.get_x() - self._padding_left - (self._border_width if self._border_left else Decimal(0)),
            cbox.get_y() - self._padding_bottom - (self._border_width if self._border_bottom else Decimal(0)),
            cbox.get_width() + self._padding_left + self._padding_right + horizontal_border_width,
            cbox.get_height() + self._padding_top + self._padding_bottom + vertical_border_width,
        )
        # fmt: on
        return self._previous_layout_box

    def paint(self, page: "Page", available_space: Rectangle) -> None:  # type: ignore[name-defined]
        """
        This method paints this LayoutElement on the given Page, in the available space
        :param page:                the Page on which to paint this LayoutElement
        :param available_space:     the available space (as a Rectangle) on which to paint this LayoutElement
        :return:                    None
        """
        horizontal_border_width: Decimal = Decimal(0)
        if self._border_left:
            horizontal_border_width += self._border_width
        if self._border_right:
            horizontal_border_width += self._border_width

        vertical_border_width: Decimal = Decimal(0)
        if self._border_top:
            vertical_border_width += self._border_width
        if self._border_bottom:
            vertical_border_width += self._border_width

        cbox_available_space: Rectangle = Rectangle(
            available_space.get_x()
            + self._padding_left
            + (self._border_width if self._border_left else Decimal(0)),
            available_space.get_y()
            + self._padding_bottom
            + (self._border_width if self._border_bottom else Decimal(0)),
            max(
                Decimal(0),
                available_space.get_width()
                - self._padding_left
                - self._padding_right
                - horizontal_border_width,
            ),
            max(
                Decimal(0),
                available_space.get_height()
                - self._padding_top
                - self._padding_bottom
                - vertical_border_width,
            ),
        )

        # determine content_box
        cbox: Rectangle = self._get_content_box(cbox_available_space)

        # take into account vertical_alignment
        delta_x: Decimal = Decimal(0)
        delta_y: Decimal = Decimal(0)
        if self._vertical_alignment == Alignment.MIDDLE:
            delta_y = (cbox_available_space.get_height() - cbox.get_height()) / Decimal(
                2
            )
            cbox.y -= delta_y
        if self._vertical_alignment == Alignment.BOTTOM:
            delta_y = cbox_available_space.get_height() - cbox.get_height()
            cbox.y -= delta_y

        # take into account horizontal_alignment
        if self._horizontal_alignment == Alignment.CENTERED:
            delta_x = (cbox_available_space.get_width() - cbox.get_width()) / Decimal(2)
            cbox.x += delta_x
        if self._horizontal_alignment == Alignment.RIGHT:
            delta_x = cbox_available_space.get_width() - cbox.get_width()
            cbox.x += delta_x

        # paint the background first
        bgbox: Rectangle = Rectangle(
            cbox.get_x()
            - self._padding_left
            - (self._border_width if self._border_left else Decimal(0)),
            cbox.get_y()
            - self._padding_bottom
            - (self._border_width if self._border_bottom else Decimal(0)),
            cbox.get_width()
            + self._padding_left
            + self._padding_right
            + horizontal_border_width,
            cbox.get_height()
            + self._padding_top
            + self._padding_bottom
            + vertical_border_width,
        )
        self._paint_background(page, bgbox)

        # paint the borders
        self._paint_borders(page, bgbox)

        # paint the actual content
        self._paint_content_box(page, cbox)

        # set bounding box
        self._previous_paint_box = bgbox

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height(),
            Decimal(0),
            Decimal(0),
        )

    def _paint_content_box(self, page: "Page", content_box: Rectangle) -> None:  # type: ignore[name-defined]
        pass

    def _get_border_outline(
        self, border_box: Rectangle
    ) -> typing.List[typing.Optional[typing.Tuple[Decimal, Decimal]]]:
        n: int = 0
        xll: Decimal = round(border_box.get_x(), n)
        yll: Decimal = round(border_box.get_y(), n)
        xur: Decimal = round(border_box.get_x() + border_box.get_width(), n)
        yur: Decimal = round(border_box.get_y() + border_box.get_height(), n)

        # top left arc
        points: typing.List[typing.Optional[typing.Tuple[Decimal, Decimal]]] = []
        if self._border_top and self._border_left and self._border_radius_top_left != 0:
            points += [
                (xll, yur - self._border_radius_top_left)
            ] + BlobFactory.smooth_closed_polygon(
                [
                    (xll, yur - self._border_radius_top_left),
                    (xll, yur),
                    (xll + self._border_radius_top_left, yur),
                ],
                2,
            )[
                :-6
            ]
        if self._border_left and self._border_radius_top_left == 0:
            points += [(xll, yur - self._border_radius_top_left)]
            points += [(xll, yur)]
        if self._border_top and self._border_radius_top_left == 0:
            points += [(xll + self._border_radius_top_left, yur)]

        # top
        if self._border_top:
            points += [(xur - self._border_radius_top_right, yur)]
        else:
            points += [None]

        # top right arc
        if (
            self._border_top
            and self._border_right
            and self._border_radius_top_right != 0
        ):
            points += BlobFactory.smooth_closed_polygon(
                [
                    (xur - self._border_radius_top_right, yur),
                    (xur, yur),
                    (xur, yur - self._border_radius_top_right),
                ],
                2,
            )[:-6]
        if self._border_top and self._border_radius_top_right == 0:
            points += [(xur, yur)]
        if self._border_right and self._border_radius_top_right == 0:
            points += [(xur, yur - self._border_radius_top_right)]

        # right
        if self._border_right:
            points += [(xur, yll + self._border_radius_bottom_right)]
        else:
            points += [None]

        # bottom right arc
        if (
            self._border_bottom
            and self._border_right
            and self._border_radius_bottom_right != 0
        ):
            points += BlobFactory.smooth_closed_polygon(
                [
                    (xur, yll + self._border_radius_bottom_right),
                    (xur, yll),
                    (xur - self._border_radius_bottom_right, yll),
                ],
                2,
            )[:-6]
        if self._border_right and self._border_radius_bottom_right == 0:
            points += [(xur, yll)]
        if self._border_bottom and self._border_radius_bottom_right == 0:
            points += [(xur - self._border_radius_bottom_right, yll)]

        # bottom
        if self._border_bottom:
            points += [(xll + self._border_radius_bottom_left, yll)]
        else:
            points += [None]

        # bottom left arc
        if (
            self._border_bottom
            and self._border_left
            and self._border_radius_bottom_left != 0
        ):
            points += BlobFactory.smooth_closed_polygon(
                [
                    (xll + self._border_radius_bottom_left, yll),
                    (xll, yll),
                    (xll, yll + self._border_radius_bottom_left),
                ],
                2,
            )[:-6]
        if self._border_bottom and self._border_radius_bottom_left == 0:
            points += [(xll, yll)]
        if self._border_left and self._border_radius_bottom_left == 0:
            points += [(xll, yll + self._border_radius_bottom_right)]

        # left
        if self._border_left:
            points += [(xll, yur - self._border_radius_top_left)]
        else:
            points += [None]

        # return
        return points

    def _paint_background(
        self, page: "Page", background_box: Rectangle  # type: ignore[name-defined]
    ):
        if not self._background_color:
            return
        assert self._background_color
        rgb_color = self._background_color.to_rgb()

        # easy case
        if (
            self._border_radius_top_right == 0
            and self._border_radius_top_left == 0
            and self._border_radius_bottom_left == 0
            and self._border_radius_bottom_right == 0
        ):

            # fmt: off
            content = """
                q %f %f %f rg %f %f m
                %f %f l
                %f %f l
                %f %f l
                %f %f l
                f
                Q
                """ % (
                float(rgb_color.red),
                float(rgb_color.green),
                float(rgb_color.blue),
                background_box.get_x(),                                 # ul_x
                background_box.get_y() + background_box.get_height(),   # ul_y

                background_box.get_x() + background_box.get_width(),    # ur_x
                background_box.get_y() + background_box.get_height(),   # ur_y

                background_box.get_x() + background_box.get_width(),    # lr_x
                background_box.get_y(),                                 # lr_y

                background_box.get_x(),                                 # ll_x
                background_box.get_y(),                                 # ll_y

                background_box.get_x(),                                 # ul_x
                background_box.get_y() + background_box.get_height(),   # ul_y
            )
            # fmt: on
            page._append_to_content_stream(content)
            return

        # remember border state
        before = [
            self._border_top,
            self._border_right,
            self._border_bottom,
            self._border_left,
        ]

        # set all borders
        self._border_top = True
        self._border_right = True
        self._border_bottom = True
        self._border_left = True

        # get outline
        outline_points = self._get_border_outline(background_box)
        assert outline_points[0] is not None

        # restore all borders
        self._border_top = before[0]
        self._border_right = before[1]
        self._border_bottom = before[2]
        self._border_left = before[3]

        # write
        content = """
            q %f %f %f rg %f %f m
            """ % (
            float(rgb_color.red),
            float(rgb_color.green),
            float(rgb_color.blue),
            float(outline_points[0][0]),
            float(outline_points[0][1]),
        )
        for p in outline_points:
            assert p is not None
            content += " %f %f l" % (float(p[0]), float(p[1]))
        content += " f Q"
        page._append_to_content_stream(content)

    def _paint_borders(self, page: "Page", border_box: Rectangle):  # type: ignore[name-defined]
        # border is not wanted on any side
        if (
            self._border_top
            == self._border_right
            == self._border_bottom
            == self._border_left
            == False
        ):
            return

        # border width is set to zero
        if self._border_width == 0:
            return

        # draw border(s)
        rgb_color = self._border_color.to_rgb()
        content = "q %f %f %f RG %f w " % (
            float(rgb_color.red),
            float(rgb_color.green),
            float(rgb_color.blue),
            float(self._border_width),
        )

        # turn points into lines
        points = self._get_border_outline(border_box)
        for i, p in enumerate(points[:-1]):
            p0: typing.Optional[typing.Tuple[Decimal, Decimal]] = p
            p1: typing.Optional[typing.Tuple[Decimal, Decimal]] = points[i + 1]
            if p0 is None or p1 is None:
                continue
            content += " %d %d m %d %d l s" % (
                float(p0[0]),
                float(p0[1]),
                float(p1[0]),
                float(p1[1]),
            )
        content += " Q"
        page._append_to_content_stream(content)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
A rubber stamp annotation (PDF 1.3) displays text or graphics intended to look as if they were stamped on the
page with a rubber stamp. When opened, it shall display a pop-up window containing the text of the associated
note. Table 181 shows the annotation dictionary entries specific to this type of annotation.
"""
import enum
import typing

from borb.io.read.types import Name, Decimal as bDecimal
from borb.pdf.canvas.color.color import Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class RubberStampAnnotationIconType(enum.Enum):
    """
    This Enum represents all possible rubber stamp annotation icons
    """

    APPROVED = Name("Approved")
    AS_IS = Name("AsIs")
    CONFIDENTIAL = Name("Confidential")
    DEPARTMENTAL = Name("Departmental")
    DRAFT = Name("Draft")
    EXPERIMENTAL = Name("Experimental")
    EXPIRED = Name("Expired")
    FINAL = Name("Final")
    FOR_COMMENT = Name("ForComment")
    FOR_PUBLIC_RELEASE = Name("ForPublicRelease")
    NOT_APPROVED = Name("NotApproved")
    NOT_FOR_PUBLIC_RELEASE = Name("NotForPublicRelease")
    SOLD = Name("Sold")
    TOP_SECRET = Name("TopSecret")


class RubberStampAnnotation(Annotation):
    """
    A rubber stamp annotation (PDF 1.3) displays text or graphics intended to look as if they were stamped on the
    page with a rubber stamp. When opened, it shall display a pop-up window containing the text of the associated
    note. Table 181 shows the annotation dictionary entries specific to this type of annotation.
    """

    def __init__(
        self,
        bounding_box: Rectangle,
        name: RubberStampAnnotationIconType = RubberStampAnnotationIconType.DRAFT,
        contents: typing.Optional[str] = None,
        color: typing.Optional[Color] = None,
    ):
        super(RubberStampAnnotation, self).__init__(
            bounding_box=bounding_box, contents=contents, color=color
        )

        # (Required) The type of annotation that this dictionary describes; shall be
        # Stamp for a rubber stamp annotation.
        self[Name("Subtype")] = Name("Stamp")

        # (Optional) The name of an icon that shall be used in displaying the annotation. Conforming readers shall provide predefined icon
        # appearances for at least the following standard names:
        # Approved, Experimental, NotApproved, AsIs,
        # Expired, NotForPublicRelease, Confidential, Final, Sold,
        # Departmental, ForComment, TopSecret, Draft, ForPublicRelease
        # Additional names may be supported as well. Default value: Draft.
        # The annotation dictionary’s AP entry, if present, shall take precedence
        # over the Name entry; see Table 168 and 12.5.5, “Appearance Streams.”
        self[Name("Name")] = name.value

        # (Optional; PDF 1.4) The constant opacity value that shall be used in
        # painting the annotation (see Sections 11.2, “Overview of Transparency,”
        # and 11.3.7, “Shape and Opacity Computations”). This value shall apply to
        # all visible elements of the annotation in its closed state (including its
        # background and border) but not to the pop-up window that appears when
        # the annotation is opened.
        self[Name("CA")] = bDecimal(1)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of EventListener extracts all lines of text from a PDF Document
"""
import typing
from functools import cmp_to_key

from borb.datastructure.disjoint_set import disjointset
from borb.io.read.types import Decimal
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.chunk_of_text_render_event import (
    ChunkOfTextRenderEvent,
    LeftToRightComparator,
)
from borb.pdf.canvas.event.end_page_event import EndPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement
from borb.pdf.canvas.layout.text.line_of_text import LineOfText
from borb.pdf.page.page import Page


class SimpleLineOfTextExtraction(EventListener):
    """
    This implementation of EventListener extracts all lines of text from a PDF Document
    """

    def __init__(self):
        self._chunks_of_text: typing.List[ChunkOfTextRenderEvent] = []
        self._current_page_number: int = -1
        self._current_page: typing.Optional[Page] = None
        self._lines_of_text_per_page: typing.Dict[int, typing.List[LineOfText]] = {}

    def _event_occurred(self, event: Event) -> None:
        if isinstance(event, ChunkOfTextRenderEvent):
            self._chunks_of_text.append(event)
        if isinstance(event, BeginPageEvent):
            self._current_page = event.get_page()
            self._current_page_number += 1
            self._chunks_of_text = []
        if isinstance(event, EndPageEvent):
            self._end_page(event.get_page())

    def get_lines_of_text_for_page(self, page: int) -> typing.List[LineOfText]:
        """
        This function returns the lines of text on a given page
        """
        return self._lines_of_text_per_page.get(page, [])

    def _end_page(self, page: Page):

        # build initial disjointset
        chunks_of_text_disjoint_set = disjointset()
        for x in self._chunks_of_text:
            chunks_of_text_disjoint_set.add(x)

        # merge all ChunkOfText objects that represent a line of text
        for c0 in chunks_of_text_disjoint_set:
            for c1 in chunks_of_text_disjoint_set:
                if c0 == c1:
                    continue
                r0 = c0._baseline_bounding_box
                r1 = c1._baseline_bounding_box
                if r0.y != r1.y:
                    continue
                gap = max(r1.x - (r0.x + r0.width), r0.x - (r1.x + r1.width))
                space_gap = (
                    c0.get_space_character_width_estimate_in_user_space()
                    if r0.x < r1.x
                    else c1.get_space_character_width_estimate_in_user_space()
                )
                if gap < space_gap * Decimal(2):
                    chunks_of_text_disjoint_set.union(c0, c1)

        # combine partitions into LineOfText objects
        lines_of_text: typing.List[LineOfText] = []
        for chunks_of_text_partition in chunks_of_text_disjoint_set.sets():
            # sort
            chunks_of_text: typing.List[ChunkOfTextRenderEvent] = [
                x for x in chunks_of_text_partition
            ]
            sorted(chunks_of_text, key=cmp_to_key(LeftToRightComparator.cmp))

            # determine text
            txt = ""
            for i, c in enumerate(chunks_of_text):
                if i == 0:
                    txt += c._text
                    continue
                # decide whether we need to insert space or not
                gap = c._baseline_bounding_box.x - (
                    chunks_of_text[i - 1]._baseline_bounding_box.x
                    + chunks_of_text[i - 1]._baseline_bounding_box.width
                )
                space_gap = chunks_of_text[
                    i - 1
                ].get_space_character_width_estimate_in_user_space()
                if gap > space_gap:
                    txt += " "
                txt += c._text

            # build LineOfText
            l: LayoutElement = LineOfText(
                text=txt,
                font=chunks_of_text[0]._font,
                font_size=chunks_of_text[0]._font_size,
                font_color=chunks_of_text[0]._font_color,
            )
            l._previous_layout_box = Rectangle(
                chunks_of_text[0]._baseline_bounding_box.x,
                chunks_of_text[0]._baseline_bounding_box.y,
                chunks_of_text[-1]._baseline_bounding_box.x
                + chunks_of_text[-1]._baseline_bounding_box.width
                - chunks_of_text[0]._baseline_bounding_box.x,
                chunks_of_text[0]._baseline_bounding_box.height,
            )
            assert isinstance(l, LineOfText)
            lines_of_text.append(l)

        # add to dict
        self._lines_of_text_per_page[self._current_page_number] = lines_of_text



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of FormField represents a drop down list that is pre-filled with countries.
"""
import typing

from borb.io.read.types import Decimal
from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.layout.forms.drop_down_list import DropDownList


class CountryDropDownList(DropDownList):
    """
    This implementation of FormField represents a drop down list that is pre-filled with countries.
    """

    COUNTRIES: typing.List[str] = [
        "Afghanistan",
        "Albania",
        "Algeria",
        "Andorra",
        "Angola",
        "Antigua and Barbuda",
        "Argentina",
        "Armenia",
        "Australia",
        "Austria",
        "Azerbaijan",
        "Bahamas",
        "Bahrain",
        "Bangladesh",
        "Barbados",
        "Belarus",
        "Belgium",
        "Belize",
        "Benin",
        "Bhutan",
        "Bolivia",
        "Bosnia and Herzegovina",
        "Botswana",
        "Brazil",
        "Brunei",
        "Bulgaria",
        "Burkina Faso",
        "Burundi",
        "Côte d'Ivoire",
        "Cabo Verde",
        "Cambodia",
        "Cameroon",
        "Canada",
        "Central African Republic",
        "Chad",
        "Chile",
        "China",
        "Colombia",
        "Comoros",
        "Congo",
        "Costa Rica",
        "Croatia",
        "Cuba",
        "Cyprus",
        "Czechia",
        "Democratic Republic of the Congo",
        "Denmark",
        "Djibouti",
        "Dominica",
        "Dominican Republic",
        "Ecuador",
        "Egypt",
        "El Salvador",
        "Equatorial Guinea",
        "Eritrea",
        "Estonia",
        "Eswatini",
        "Ethiopia",
        "Fiji",
        "Finland",
        "France",
        "Gabon",
        "Gambia",
        "Georgia",
        "Germany",
        "Ghana",
        "Greece",
        "Grenada",
        "Guatemala",
        "Guinea",
        "Guinea-Bissau",
        "Guyana",
        "Haiti",
        "Holy See",
        "Honduras",
        "Hungary",
        "Iceland",
        "India",
        "Indonesia",
        "Iran",
        "Iraq",
        "Ireland",
        "Israel",
        "Italy",
        "Jamaica",
        "Japan",
        "Jordan",
        "Kazakhstan",
        "Kenya",
        "Kiribati",
        "Kuwait",
        "Kyrgyzstan",
        "Laos",
        "Latvia",
        "Lebanon",
        "Lesotho",
        "Liberia",
        "Libya",
        "Liechtenstein",
        "Lithuania",
        "Luxembourg",
        "Madagascar",
        "Malawi",
        "Malaysia",
        "Maldives",
        "Mali",
        "Malta",
        "Marshall Islands",
        "Mauritania",
        "Mauritius",
        "Mexico",
        "Micronesia",
        "Moldova",
        "Monaco",
        "Mongolia",
        "Montenegro",
        "Morocco",
        "Mozambique",
        "Myanmar",
        "Namibia",
        "Nauru",
        "Nepal",
        "Netherlands",
        "New Zealand",
        "Nicaragua",
        "Niger",
        "Nigeria",
        "North Korea",
        "North Macedonia",
        "Norway",
        "Oman",
        "Pakistan",
        "Palau",
        "Palestine State",
        "Panama",
        "Papua New Guinea",
        "Paraguay",
        "Peru",
        "Philippines",
        "Poland",
        "Portugal",
        "Qatar",
        "Romania",
        "Russia",
        "Rwanda",
        "Saint Kitts and Nevis",
        "Saint Lucia",
        "Saint Vincent and the Grenadines",
        "Samoa",
        "San Marino",
        "Sao Tome and Principe",
        "Saudi Arabia",
        "Senegal",
        "Serbia",
        "Seychelles",
        "Sierra Leone",
        "Singapore",
        "Slovakia",
        "Slovenia",
        "Solomon Islands",
        "Somalia",
        "South Africa",
        "South Korea",
        "South Sudan",
        "Spain",
        "Sri Lanka",
        "Sudan",
        "Suriname",
        "Sweden",
        "Switzerland",
        "Syria",
        "Tajikistan",
        "Tanzania",
        "Thailand",
        "Timor-Leste",
        "Togo",
        "Tonga",
        "Trinidad and Tobago",
        "Tunisia",
        "Turkey",
        "Turkmenistan",
        "Tuvalu",
        "Uganda",
        "Ukraine",
        "United Arab Emirates",
        "United Kingdom",
        "United States of America",
        "Uruguay",
        "Uzbekistan",
        "Vanuatu",
        "Venezuela",
        "Vietnam",
        "Yemen",
        "Zambia",
        "Zimbabwe",
    ]

    def __init__(
        self,
        default_value: str = "",
        field_name: typing.Optional[str] = None,
        font_color: Color = HexColor("000000"),
        font_size: Decimal = Decimal(12),
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_top: typing.Optional[Decimal] = None,
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        value: str = "",
    ):
        super(CountryDropDownList, self).__init__(
            default_value=default_value,
            field_name=field_name,
            font_color=font_color,
            font_size=font_size,
            margin_bottom=margin_bottom,
            margin_left=margin_left,
            margin_right=margin_right,
            margin_top=margin_top,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            possible_values=CountryDropDownList.COUNTRIES,
            value=value,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class represents a line segment
"""
from decimal import Decimal
from math import sqrt
from typing import Tuple

from borb.pdf.canvas.geometry.matrix import Matrix


class LineSegment:
    """
    This class represents a line segment
    """

    def __init__(self, x0: Decimal, y0: Decimal, x1: Decimal, y1: Decimal):
        self.x0: Decimal = x0
        self.y0: Decimal = y0
        self.x1: Decimal = x1
        self.y1: Decimal = y1

    def length(self) -> Decimal:
        """
        This function returns the length of this LineSegment
        """
        return Decimal(sqrt((self.x0 - self.x1) ** 2 + (self.y0 - self.y1) ** 2))

    def get_start(self) -> Tuple[Decimal, Decimal]:
        """
        This function returns the start of this LineSegment
        """
        return (self.x0, self.y0)

    def get_end(self) -> Tuple[Decimal, Decimal]:
        """
        This function returns the end of this LineSegment
        """
        return (self.x1, self.y1)

    def transform_by(self, matrix: Matrix) -> "LineSegment":
        """
        This function transforms the start and end of this LineSegment by a given Matrix,
        it returns the transformed LineSegment
        """
        p0 = matrix.cross(self.x0, self.y0, Decimal(1))
        p1 = matrix.cross(self.x1, self.y1, Decimal(1))
        return LineSegment(p0[0], p0[1], p1[0], p1[1])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    In cryptography, RC4 (Rivest Cipher 4 also known as ARC4 or ARCFOUR meaning Alleged RC4, see below) is a stream cipher.
    While it is remarkable for its simplicity and speed in software, multiple vulnerabilities have been discovered in RC4, rendering it insecure.
    It is especially vulnerable when the beginning of the output keystream is not discarded, or when nonrandom or related keys are used.
    Particularly problematic uses of RC4 have led to very insecure protocols such as WEP.

    As of 2015, there is speculation that some state cryptologic agencies may possess the capability to break RC4 when used in the TLS protocol.
    IETF has published RFC 7465 to prohibit the use of RC4 in TLS; Mozilla and Microsoft have issued similar recommendations.

    A number of attempts have been made to strengthen RC4, notably Spritz, RC4A, VMPC, and RC4+.
"""


class RC4:
    """
    In cryptography, RC4 (Rivest Cipher 4 also known as ARC4 or ARCFOUR meaning Alleged RC4, see below) is a stream cipher.
    While it is remarkable for its simplicity and speed in software, multiple vulnerabilities have been discovered in RC4, rendering it insecure.
    It is especially vulnerable when the beginning of the output keystream is not discarded, or when nonrandom or related keys are used.
    Particularly problematic uses of RC4 have led to very insecure protocols such as WEP.

    As of 2015, there is speculation that some state cryptologic agencies may possess the capability to break RC4 when used in the TLS protocol.
    IETF has published RFC 7465 to prohibit the use of RC4 in TLS; Mozilla and Microsoft have issued similar recommendations.

    A number of attempts have been made to strengthen RC4, notably Spritz, RC4A, VMPC, and RC4+.
    """

    def __init__(self):
        self._state = [n for n in range(256)]
        self._p: int = 0
        self._q: int = 0

    def _set_key(self, key: bytes):
        self._state = [n for n in range(256)]
        self._p = 0
        self._q = 0
        j: int = 0
        for i in range(256):
            if len(key) > 0:
                j = (j + self._state[i] + key[i % len(key)]) % 256
            else:
                j = (j + self._state[i]) % 256
            self._state[i], self._state[j] = self._state[j], self._state[i]

    def _byte_generator(self):
        self._p = (self._p + 1) % 256
        self._q = (self._q + self._state[self._p]) % 256
        self._state[self._p], self._state[self._q] = (
            self._state[self._q],
            self._state[self._p],
        )
        return self._state[(self._state[self._p] + self._state[self._q]) % 256]

    def encrypt(self, key: bytes, input: bytes):
        """
        This function encrypts a given byte array with a given key,
        returning the encrypted bytes.
        :param key:     the key to be used for encrypting
        :param input:   the input byte array to be encrypted
        :return:        the encrypted bytes
        """
        self._set_key(key)
        return bytes([p ^ self._byte_generator() for p in input])



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Move to the start of the next line, offset from the start of the current line by
(tx , ty). t x and t y shall denote numbers expressed in unscaled text space
units. More precisely, this operator shall perform these assignments:
Tm = Tlm = [[1,0,0], [0,1,0],[tx,ty,1]] * Tlm
"""
import copy
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.geometry.matrix import Matrix
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class MoveTextPosition(CanvasOperator):
    """
    Move to the start of the next line, offset from the start of the current line by
    (tx , ty). t x and t y shall denote numbers expressed in unscaled text space
    units. More precisely, this operator shall perform these assignments:
    Tm = Tlm = [[1,0,0], [0,1,0],[tx,ty,1]] * Tlm
    """

    def __init__(self):
        super().__init__("Td", 2)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the Td operator
        """
        assert isinstance(operands[0], Decimal), "Operand 0 of Td must be a Decimal"
        assert isinstance(operands[1], Decimal), "Operand 1 of Td must be a Decimal"

        tx = operands[0]
        ty = operands[1]

        m = Matrix.identity_matrix()
        m[2][0] = tx
        m[2][1] = ty

        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.text_matrix = m.mul(
            canvas.graphics_state.text_line_matrix
        )
        canvas.graphics_state.text_line_matrix = copy.deepcopy(
            canvas.graphics_state.text_matrix
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This class was meant to perform a dictionary/array level comparison of PDF documents.
    It makes it a lot easier to debug problems.
"""
import typing

from borb.io.filter.stream_decode_util import decode_stream
from borb.io.read.types import Decimal, Dictionary, List, Name, Stream
from borb.pdf.document.document import Document


class PDFDiff:
    """
    This class was meant to perform a dictionary/array level comparison of PDF documents.
    It makes it a lot easier to debug problems.
    """

    def __init__(self, pdf_a: Document, pdf_b: Document):
        self._document_a: Document = pdf_a
        self._document_b: Document = pdf_b
        self._already_compared: typing.List[int] = []
        self._errors: typing.List[str] = []

    def compare(self) -> None:
        """
        This method compares the given PDF documents, logging any differences between them.
        """
        self._compare(self._document_a, self._document_b, "", "")

    @staticmethod
    def _get_reference_or_none(obj) -> str:
        try:
            if obj.get_reference() is not None:
                return "(%d 0 R)" % obj.get_reference().object_number
        except:
            pass
        return ""

    def _log_difference(self, error_msg: str) -> None:
        print(error_msg)
        self._errors.append(error_msg)

    def _compare(self, a, b, path_to_a, path_to_b) -> None:

        if id(a) in self._already_compared:
            return
        if id(b) in self._already_compared:
            return
        self._already_compared.append(id(a))
        self._already_compared.append(id(b))

        # check type
        if a.__class__.__name__ != b.__class__.__name__:
            self._log_difference(
                "Class mismatch : %s %s <--> %s %s"
                % (path_to_a, a.__class__.__name__, path_to_b, b.__class__.__name__)
            )

        if isinstance(a, Name):
            if str(a) != str(b):
                self._log_difference(
                    "Name mismatch : %s %s <--> %s %s"
                    % (path_to_a, str(a), path_to_b, str(b))
                )
            return

        if isinstance(a, Decimal):
            if int(a) != int(b):
                self._log_difference(
                    "Value mismatch : %s %s <--> %s %s"
                    % (path_to_a, str(a), path_to_b, str(b))
                )

        # get references if they exist
        ref_a = PDFDiff._get_reference_or_none(a)
        ref_b = PDFDiff._get_reference_or_none(a)

        # compare streams
        if isinstance(a, Stream):
            decode_stream(a)
            decode_stream(b)
            if "DecodedBytes" not in a:
                self._log_difference("Unable to decode Stream %s" % (path_to_a + ref_a))
            if "DecodedBytes" not in b:
                self._log_difference("Unable to decode Stream %s" % (path_to_b + ref_b))
            dba: bytes = a["DecodedBytes"]
            dbb: bytes = b["DecodedBytes"]
            if len(dba) != len(dbb):
                self._errors.append(
                    "Stream Length mismatch : %s %d <--> %s %d"
                    % (path_to_a + ref_a, len(a), path_to_b + ref_b, len(b))
                )
            else:
                for i in range(0, len(dba)):
                    if dba[i] != dbb[i]:
                        self._errors.append(
                            "Stream content mismatch : %s %d <--> %s %d"
                            % (path_to_a + ref_a, i, path_to_b + ref_b, i)
                        )

        # compare dictionary
        if isinstance(a, Dictionary):
            for k, v in a.items():
                if k == "ID":
                    continue
                if k == "Bytes":
                    continue
                if k == "DecodedBytes":
                    continue
                if isinstance(a, Stream) and k == "Length":
                    continue
                if k not in b:
                    self._log_difference(
                        "Key absent/present mismatch : %s %s <--> %s %s"
                        % (path_to_a + ref_a, str(k), path_to_b + ref_b, None)
                    )
                    continue
                self._compare(
                    a[k],
                    b[k],
                    path_to_a + "/" + str(k) + ref_a,
                    path_to_b + "/" + str(k) + ref_b,
                )
            return

        # compare array
        if isinstance(a, List):
            if len(a) != len(b):
                self._errors.append(
                    "Array Length mismatch : %s %d <--> %s %d"
                    % (path_to_a + ref_a, len(a), path_to_b + ref_b, len(b))
                )
            for i in range(0, min(len(a), len(b))):
                self._compare(
                    a[i],
                    b[i],
                    path_to_a + ref_a + "/" + str(i),
                    path_to_b + ref_b + "/" + str(i),
                )
            return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Polygon annotations (PDF 1.5) display closed polygons on the page. Such polygons may have any number of
vertices connected by straight lines. Polyline annotations (PDF 1.5) are similar to polygons, except that the first
and last vertex are not implicitly connected.
"""
import enum
import typing
from decimal import Decimal

from borb.io.read.types import Name, List, Decimal as bDecimal
from borb.pdf.canvas.color.color import HexColor, Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class LineEndStyleType(enum.Enum):
    """
    This Enum represents all possible line end styles
    """

    SQUARE = Name("Square")
    CIRCLE = Name("Circle")
    DIAMOND = Name("Diamond")
    OPEN_ARROW = Name("OpenArrow")
    CLOSED_ARROW = Name("ClosedArrow")
    NONE = Name("None")
    BUTT = Name("Butt")
    RIGHT_OPEN_ARROW = Name("ROpenArrow")
    RIGHT_CLOSED_ARROW = Name("RClosedArrow")
    SLASH = Name("Slash")


class PolylineAnnotation(Annotation):
    """
    Polygon annotations (PDF 1.5) display closed polygons on the page. Such polygons may have any number of
    vertices connected by straight lines. Polyline annotations (PDF 1.5) are similar to polygons, except that the first
    and last vertex are not implicitly connected.
    """

    def __init__(
        self,
        points: typing.List[typing.Tuple[Decimal, Decimal]],
        stroke_color: typing.Optional[Color] = HexColor("000000"),
        fill_color: typing.Optional[Color] = None,
        left_line_end_style: LineEndStyleType = LineEndStyleType.NONE,
        right_line_end_style: LineEndStyleType = LineEndStyleType.NONE,
    ):

        # must be at least 3 points
        assert len(points) >= 3

        # bounding box
        min_x = points[0][0]
        min_y = points[0][1]
        max_x = min_x
        max_y = min_y
        for p in points:
            min_x = min(min_x, p[0])
            min_y = min(min_y, p[1])
            max_x = max(max_x, p[0])
            max_y = max(max_y, p[1])

        # create generic annotation
        super(PolylineAnnotation, self).__init__(
            bounding_box=Rectangle(min_x, min_y, max_x - min_x, max_y - min_y),
            color=stroke_color,
        )

        # (Required) The type of annotation that this dictionary describes; shall be
        # Polygon or PolyLine for a polygon or polyline annotation, respectively.
        self[Name("Subtype")] = Name("PolyLine")

        # (Required) An array of numbers (see Table 174) specifying the width and
        # dash pattern that shall represent the alternating horizontal and vertical
        # coordinates, respectively, of each vertex, in default user space.
        self[Name("Vertices")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        for p in points:
            self["Vertices"].append(bDecimal(p[0]))
            self["Vertices"].append(bDecimal(p[1]))

        # (Optional; PDF 1.4) An array of two names specifying the line ending
        # styles that shall be used in drawing the line. The first and second
        # elements of the array shall specify the line ending styles for the endpoints
        # defined, respectively, by the first and second pairs of coordinates, (x 1 , y 1 )
        # and (x 2 , y 2 ), in the L array. Table 176 shows the possible values. Default
        # value: [ /None /None ].
        self[Name("LE")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self["LE"].append(left_line_end_style)
        self["LE"].append(right_line_end_style)

        if fill_color is not None:
            self[Name("IC")] = List().set_is_inline(True)  # type: ignore [attr-defined]
            self["IC"].append(bDecimal(fill_color.to_rgb().red))
            self["IC"].append(bDecimal(fill_color.to_rgb().green))
            self["IC"].append(bDecimal(fill_color.to_rgb().blue))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    This implementation of Event is triggered right before the Canvas is being processed.
"""
from borb.pdf.canvas.event.event_listener import Event
from borb.pdf.page.page import Page


class BeginPageEvent(Event):
    """
    This implementation of Event is triggered right before the Canvas is being processed.
    """

    def __init__(self, page: Page):
        self._page: Page = page

    def get_page(self) -> Page:
        """
        This function returns the Page that triggered this BeginPageEvent
        """
        return self._page



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of EventListener resizes all Image objects on a Page
to fit their actual dimensions (ensuring they are not bigger than they need to be)
"""
import typing

from PIL import Image as PILImage  # type: ignore [import]

from borb.io.read.types import Name, add_base_methods
from borb.pdf.canvas.event.begin_page_event import BeginPageEvent
from borb.pdf.canvas.event.event_listener import Event, EventListener
from borb.pdf.canvas.event.image_render_event import ImageRenderEvent
from borb.pdf.page.page import Page


class ImageFormatOptimization(EventListener):
    """
    This implementation of EventListener resizes all Image objects on a Page
    to fit their actual dimensions (ensuring they are not bigger than they need to be)
    """

    def __init__(self):
        super(ImageFormatOptimization, self).__init__()
        self._current_page: typing.Optional[Page] = None

    def _event_occurred(self, event: "Event") -> None:
        if isinstance(event, BeginPageEvent):
            self._current_page = event.get_page()
        if isinstance(event, ImageRenderEvent):
            self._render_image(event)

    def _render_image(self, image_render_event: "ImageRenderEvent"):
        source_image: PILImage = image_render_event.get_image()  # type: ignore[valid-type]

        # get desired width/height
        w0: int = int(image_render_event.get_width())  # type: ignore [attr-defined]
        h0: int = int(image_render_event.get_height())  # type: ignore [attr-defined]

        # get current width/height
        w1: int = source_image.width  # type: ignore[attr-defined]
        h1: int = source_image.height  # type: ignore[attr-defined]

        # get image resource name
        assert self._current_page is not None
        resource_name: typing.Optional[Name] = next(
            iter(
                [
                    k
                    for k, v in self._current_page["Resources"]["XObject"].items()
                    if id(v) == id(source_image)
                ]
            ),
            None,
        )
        assert resource_name is not None

        # resize
        if (w0 * h0) < (w1 * h1):
            resized_image: PILImage = source_image.resize((w0, h0))  # type: ignore[attr-defined, valid-type]
            add_base_methods(resized_image)
            self._current_page["Resources"]["XObject"][resource_name] = resized_image
            resized_image.set_parent(self._current_page["Resources"]["XObject"][resource_name])  # type: ignore[attr-defined]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for writing Stream objects
"""
import logging
import typing
import zlib
from typing import Optional

from borb.io.read.types import AnyPDFType
from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, List, Name, Reference, Stream
from borb.io.write.transformer import Transformer, WriteTransformerState

logger = logging.getLogger(__name__)


class StreamTransformer(Transformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing Stream objects
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be converted represents a Stream object
        """
        return isinstance(any, Stream)

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes a Stream to a byte stream
        """
        # fmt: off
        assert (context is not None), "context must be defined in order to write Stream objects."
        assert context.destination is not None, "context.destination must be defined in order to write Stream objects."
        assert isinstance(object_to_transform, Stream), "object_to_transform must be of type Stream"
        # fmt: on

        # avoid resolving objects twice
        object_ref: typing.Optional[Reference] = object_to_transform.get_reference()  # type: ignore [attr-defined]
        if object_ref is not None and object_ref in context.resolved_references:
            assert object_ref is not None
            assert object_ref.object_number is not None
            logger.debug(
                "skip writing object %d %d R (already resolved)"
                % (object_ref.object_number, object_ref.generation_number or 0)
            )
            return

        # start object if needed
        started_object = False
        if object_ref is not None:
            assert object_ref.object_number is not None
            if object_ref.object_number is not None and object_ref.byte_offset is None:
                started_object = True
                self._start_object(object_to_transform, context)
            context.resolved_references.append(object_ref)

        # build stream dictionary
        stream_dictionary = Dictionary()

        # objects to turn into reference
        queue: typing.List[AnyPDFType] = []
        for k, v in object_to_transform.items():
            if k in ["Bytes", "DecodedBytes"]:
                continue
            if (
                isinstance(v, Dictionary)
                or isinstance(v, List)
                or isinstance(v, Stream)
            ) and not v.is_inline():  # type: ignore [union-attr]
                stream_dictionary[k] = self.get_reference(v, context)
                queue.append(v)
            else:
                stream_dictionary[k] = v

        # if self.compression_level == 0, remove /Filter
        if context.compression_level == 0 and Name("Filter") in stream_dictionary:
            stream_dictionary.pop(Name("Filter"))

        # handle compression
        if "DecodedBytes" in object_to_transform:
            if context.compression_level == 0:
                bts = object_to_transform["DecodedBytes"]
            else:
                bts = zlib.compress(
                    object_to_transform["DecodedBytes"], context.compression_level
                )
            stream_dictionary[Name("Length")] = bDecimal(len(bts))
        else:
            assert "Bytes" in object_to_transform
            bts = object_to_transform["Bytes"]

        # write stream dictionary
        self.get_root_transformer().transform(stream_dictionary, context)

        # write "stream"
        context.destination.write(bytes("stream\n", "latin1"))

        # write bytes
        context.destination.write(bts)

        # write "endstream"
        context.destination.write(bytes("\nendstream\n", "latin1"))

        # end object if needed
        if started_object:
            self._end_object(object_to_transform, context)

        for e in queue:
            self.get_root_transformer().transform(e, context)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents a basic (no-text) progressbar.
It displays a rectangular shape of fill_color, overlaid with a (smaller) rectangular shape of
stroke_color. This implementation of ProgressBar is roughly size 12 font wide.
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement, Alignment


class ProgressBar(LayoutElement):
    """
    This implementation of LayoutElement represents a basic (no-text) progressbar.
    It displays a rectangular shape of fill_color, overlaid with a (smaller) rectangular shape of
    stroke_color. This implementation of ProgressBar is roughly size 12 font wide.
    """

    def __init__(
        self,
        percentage: float = 0.0,
        vertical_alignment: Alignment = Alignment.TOP,
        horizontal_alignment: Alignment = Alignment.LEFT,
        border_top: bool = False,
        border_right: bool = False,
        border_bottom: bool = False,
        border_left: bool = False,
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_bottom_left: Decimal = Decimal(0),
        border_color: Color = HexColor("000000"),
        border_width: Decimal = Decimal(1),
        padding_top: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        margin_top: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        background_color: typing.Optional[Color] = None,
        stroke_color: Color = HexColor("2c99f9"),
        fill_color: Color = HexColor("f0f0f0"),
    ):
        super().__init__(
            font_size=Decimal(0),
            border_top=border_top,
            border_right=border_right,
            border_bottom=border_bottom,
            border_left=border_left,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_bottom_left=border_radius_bottom_left,
            border_color=border_color,
            border_width=border_width,
            padding_top=padding_top,
            padding_right=padding_right,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            margin_top=margin_top or Decimal(0),
            margin_right=margin_right or Decimal(0),
            margin_bottom=margin_bottom or Decimal(0),
            margin_left=margin_left or Decimal(0),
            vertical_alignment=vertical_alignment,
            horizontal_alignment=horizontal_alignment,
            background_color=background_color,
        )
        assert 0 <= percentage <= 1, "ProgressBar displays percentages between 0 and 1"
        self._percentage: float = percentage
        self._stroke_color: Color = stroke_color
        self._fill_color: Color = fill_color

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        h: Decimal = min(available_space.get_height(), Decimal(12 * 1.2))
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - h,
            available_space.get_width(),
            h,
        )

    def _paint_content_box(self, page: "Page", content_box: Rectangle) -> None:  # type: ignore[name-defined]

        # draw rectangle background
        fill_rgb = (self._fill_color or HexColor("f0f0f0")).to_rgb()
        content = " q %f %f %f RG %f %f %f rg 0.1 w " % (
            float(fill_rgb.red),
            float(fill_rgb.green),
            float(fill_rgb.blue),
            float(fill_rgb.red),
            float(fill_rgb.green),
            float(fill_rgb.blue),
        )
        content += " %f %f %f %f re B" % (
            float(content_box.get_x()),
            float(content_box.get_y()),
            float(content_box.get_width()),
            float(content_box.get_height()),
        )

        # draw active color background
        if self._percentage != 0:
            stroke_rgb = (self._stroke_color or HexColor("2c99f9")).to_rgb()
            content += " %f %f %f RG  %f %f %f rg" % (
                float(stroke_rgb.red),
                float(stroke_rgb.green),
                float(stroke_rgb.blue),
                float(stroke_rgb.red),
                float(stroke_rgb.green),
                float(stroke_rgb.blue),
            )
            content += " %f %f %f %f re B" % (
                float(content_box.get_x()),
                float(content_box.get_y()),
                float(content_box.get_width() * Decimal(self._percentage)),
                float(content_box.get_height()),
            )

        # end stack
        content += " Q"

        # append to page
        page._append_to_content_stream(content)


class ProgressSquare(ProgressBar):
    """
    This implementation of LayoutElement represents a basic (no-text) progressbar.
    It displays a rectangular shape of fill_color, overlaid with a (smaller) rectangular shape of
    stroke_color. This implementation of ProgressBar is roughly size 12 font tall AND wide.
    """

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        h: Decimal = min(available_space.get_height(), Decimal(12 * 1.2))
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - h,
            h,
            h,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer
is responsible for writing XMP meta-data information
"""
import logging
import xml.etree.ElementTree as ET
from typing import Optional

from borb.io.read.types import AnyPDFType
from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Name, Reference, Stream
from borb.io.write.transformer import Transformer, WriteTransformerState

logger = logging.getLogger(__name__)


class XMPTransformer(Transformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing XMP meta-data information
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be converted represents an XML element
        """
        return isinstance(any, ET.Element)

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes an ET.Element (representing XMP meta information) to a byte stream
        """
        assert isinstance(object_to_transform, ET.Element)
        # fmt: off
        assert context is not None, "A WriteTransformerState must be defined in order to write XMP objects."
        assert context.destination is not None, "A WriteTransformerState must be defined in order to write XMP objects."
        # fmt: on

        # build stream
        out_value = Stream()
        out_value[Name("Type")] = Name("Metadata")
        out_value[Name("Subtype")] = Name("XML")

        bts = ET.tostring(object_to_transform)
        out_value[Name("DecodedBytes")] = bts
        out_value[Name("Bytes")] = bts
        out_value[Name("Length")] = bDecimal(len(bts))

        # copy reference
        out_value.set_reference(object_to_transform.get_reference())  # type: ignore [attr-defined]

        # start object if needed
        started_object = False
        ref = out_value.get_reference()  # type: ignore [attr-defined]
        if ref is not None:
            assert isinstance(ref, Reference)
            if ref.object_number is not None and ref.byte_offset is None:
                started_object = True
                self._start_object(out_value, context)

        # pass stream along to other transformer
        self.get_root_transformer().transform(out_value, context)

        # end object if needed
        if started_object:
            self._end_object(out_value, context)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Interactive forms (see 12.7, “Interactive Forms”) use widget annotations (PDF 1.2) to represent the appearance
of fields and to manage user interactions. As a convenience, when a field has only a single associated widget
annotation, the contents of the field dictionary (12.7.3, “Field Dictionaries”) and the annotation dictionary may
be merged into a single dictionary containing entries that pertain to both a field and an annotation.
"""
from borb.pdf.canvas.layout.annotation.annotation import Annotation


class WidgetAnnotation(Annotation):
    """
    Interactive forms (see 12.7, “Interactive Forms”) use widget annotations (PDF 1.2) to represent the appearance
    of fields and to manage user interactions. As a convenience, when a field has only a single associated widget
    annotation, the contents of the field dictionary (12.7.3, “Field Dictionaries”) and the annotation dictionary may
    be merged into a single dictionary containing entries that pertain to both a field and an annotation.
    """

    def __init__(self):
        super(WidgetAnnotation, self).__init__()
        # TODO



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement represents an Image
"""
import typing
from decimal import Decimal
from pathlib import Path
from typing import Optional

import requests
from PIL import Image as PILImage  # type: ignore [import]

from borb.io.read.types import Dictionary, Name, add_base_methods
from borb.pdf.canvas.color.color import HexColor, Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import Alignment, LayoutElement
from borb.pdf.page.page import Page


class Image(LayoutElement):
    """
    This implementation of LayoutElement represents an Image
    """

    def __init__(
        self,
        image: typing.Union[str, Path, PILImage.Image],
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        height: Optional[Decimal] = None,
        horizontal_alignment: Alignment = Alignment.LEFT,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_top: typing.Optional[Decimal] = None,
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        vertical_alignment: Alignment = Alignment.TOP,
        width: Optional[Decimal] = None,
    ):
        if isinstance(image, str):
            image = PILImage.open(
                requests.get(
                    image,
                    stream=True,
                ).raw
            )
        if isinstance(image, Path):
            image = PILImage.open(image)
        super(Image, self).__init__(
            background_color=None,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            font_size=Decimal(12),
            horizontal_alignment=horizontal_alignment,
            margin_bottom=margin_bottom if margin_bottom is not None else Decimal(5),
            margin_left=margin_left if margin_left is not None else Decimal(5),
            margin_right=margin_right if margin_right is not None else Decimal(5),
            margin_top=margin_top if margin_top is not None else Decimal(5),
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            vertical_alignment=vertical_alignment,
        )
        add_base_methods(image)
        self._image: PILImage = image  # type: ignore[valid-type]
        self._width = width or Decimal(self._image.width)
        self._height = height or Decimal(self._image.height)

    def _get_image_resource_name(self, image: PILImage, page: Page):  # type: ignore[valid-type]
        # create resources if needed
        if "Resources" not in page:
            page[Name("Resources")] = Dictionary().set_parent(page)  # type: ignore [attr-defined]
        if "XObject" not in page["Resources"]:
            page["Resources"][Name("XObject")] = Dictionary()

        # insert font into resources
        image_resource_name = [
            k for k, v in page["Resources"]["XObject"].items() if v == image
        ]
        if len(image_resource_name) > 0:
            return image_resource_name[0]
        else:
            image_index = len(page["Resources"]["XObject"]) + 1
            page["Resources"]["XObject"][Name("Im%d" % image_index)] = image
            return Name("Im%d" % image_index)

    def _get_content_box(self, available_space: Rectangle) -> Rectangle:
        return Rectangle(
            available_space.get_x(),
            available_space.get_y() + available_space.get_height() - self._height,
            self._width,
            self._height,
        )

    def _paint_content_box(self, page: Page, bounding_box: Rectangle):

        # add image to resources
        image_resource_name = self._get_image_resource_name(self._image, page)

        assert self._width is not None
        assert self._height is not None

        # write Do operator
        content = " q %f 0 0 %f %f %f cm /%s Do Q " % (
            float(self._width),
            float(self._height),
            float(bounding_box.get_x()),
            float(bounding_box.get_y() + bounding_box.get_height() - self._height),
            image_resource_name,
        )

        # write content
        page._append_to_content_stream(content)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This class represents a PDF document
"""
import typing
import zlib
from decimal import Decimal

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, List, Name, Stream, String
from borb.pdf.canvas.layout.annotation.link_annotation import DestinationType
from borb.pdf.document.name_tree import NameTree
from borb.pdf.page.page import Page
from borb.pdf.trailer.document_info import DocumentInfo, XMPDocumentInfo
from borb.pdf.xref.plaintext_xref import PlainTextXREF


class Document(Dictionary):
    """
    This class represents a PDF document
    """

    def get_document_info(self) -> DocumentInfo:
        """
        This function returns the DocumentInfo of this Document
        """
        return DocumentInfo(self)

    def get_xmp_document_info(self) -> XMPDocumentInfo:
        """
        This function returns the XMPDocumentInfo of this Document
        """
        return XMPDocumentInfo(self)

    #
    # adding/removing pages
    #

    def get_page(self, page_number: int) -> Page:
        """
        This function returns a Page (at given page_number) within this Document
        """
        return self["XRef"]["Trailer"]["Root"]["Pages"]["Kids"][page_number]

    def add_document(self, document: "Document") -> "Document":
        """
        This method appends another Document to this one
        """
        number_of_pages_in_other = int(
            document.get_document_info().get_number_of_pages() or 0
        )
        for i in range(0, number_of_pages_in_other):
            self.add_page(document.get_page(i))
        return self

    def add_page(self, page: Page) -> "Document":  # type: ignore [name-defined]
        """
        This method appends a page (from another Document) to this Document
        """
        return self.insert_page(page)

    def insert_page(self, page: Page, index: typing.Optional[int] = None) -> "Document":  # type: ignore [name-defined]
        """
        This method appends a page (from another Document) to this Document at a given index
        """
        # build XRef
        if "XRef" not in self:
            self[Name("XRef")] = PlainTextXREF()
            self[Name("XRef")].set_parent(self)
        # build Trailer
        if "Trailer" not in self["XRef"]:
            self["XRef"][Name("Trailer")] = Dictionary()
            self["XRef"][Name("Size")] = bDecimal(0)
            self["XRef"]["Trailer"].set_parent(self["XRef"])
        # build Root
        if "Root" not in self["XRef"]["Trailer"]:
            self["XRef"]["Trailer"][Name("Root")] = Dictionary()
            self["XRef"]["Trailer"]["Root"][Name("Type")] = Name("Catalog")
            self["XRef"]["Trailer"]["Root"].set_parent(self["XRef"]["Trailer"])
        # build Pages
        if "Pages" not in self["XRef"]["Trailer"]["Root"]:
            # fmt: off
            self["XRef"]["Trailer"][Name("Root")][Name("Pages")] = Dictionary()
            self["XRef"]["Trailer"][Name("Root")][Name("Pages")][Name("Count")] = bDecimal(0)
            self["XRef"]["Trailer"][Name("Root")][Name("Pages")][Name("Kids")] = List()
            self["XRef"]["Trailer"][Name("Root")][Name("Pages")][Name("Type")] = Name("Pages")
            self["XRef"]["Trailer"]["Root"]["Pages"].set_parent(self["XRef"]["Trailer"]["Root"])
            self["XRef"]["Trailer"]["Root"]["Pages"]["Kids"].set_parent(self["XRef"]["Trailer"]["Root"]["Pages"])
            # fmt: on
        # update /Kids
        kids = self["XRef"]["Trailer"]["Root"]["Pages"]["Kids"]
        assert kids is not None
        assert isinstance(kids, List)
        if index is None:
            index = len(kids)
        kids.insert(index, page)
        # update /Count
        prev_count = self["XRef"]["Trailer"]["Root"]["Pages"]["Count"]
        self["XRef"]["Trailer"]["Root"]["Pages"][Name("Count")] = bDecimal(
            prev_count + 1
        )
        # set /Parent
        page[Name("Parent")] = self["XRef"]["Trailer"]["Root"]["Pages"]
        page.set_parent(kids)  # type: ignore [attr-defined]
        # return
        return self

    def pop_page(self, index: int) -> "Document":  # type: ignore [name-specified]
        """
        This method removes a Page from this Document at a given index.
        It then returns this Document.
        """
        if "XRef" not in self:
            return self
        if "Trailer" not in self["XRef"]:
            return self
        if "Root" not in self["XRef"]["Trailer"]:
            return self
        if "Pages" not in self["XRef"]["Trailer"]["Root"]:
            return self
        if "Kids" not in self["XRef"]["Trailer"]["Root"]["Pages"]:
            return self
        if "Count" not in self["XRef"]["Trailer"]["Root"]["Pages"]:
            return self

        # get Kids
        kids = self["XRef"]["Trailer"]["Root"]["Pages"]["Kids"]
        assert kids is not None
        assert isinstance(kids, List)

        # out of bounds
        if index < 0 or index >= len(kids):
            return self

        # remove
        kids.pop(index)
        self["XRef"]["Trailer"]["Root"]["Pages"][Name("Count")] = bDecimal(len(kids))

        # return
        return self

    #
    # outlines
    #

    def has_outlines(self) -> bool:
        """
        A PDF document may contain a document outline that the conforming reader may display on the screen,
        allowing the user to navigate interactively from one part of the document to another. The outline consists of a
        tree-structured hierarchy of outline items (sometimes called bookmarks), which serve as a visual table of
        contents to display the document’s structure to the user.
        This function returns True if this Document has outlines, false otherwise
        """
        try:
            return (
                "Outlines" in self["XRef"]["Trailer"]["Root"]
                and "Count" in self["XRef"]["Trailer"]["Root"]["Outlines"]
                and self["XRef"]["Trailer"]["Root"]["Outlines"]["Count"] > 0
            )
        except:
            return False

    def add_outline(
        self,
        text: str,
        level: int,
        destination_type: DestinationType,
        page_nr: int,
        top: typing.Optional[Decimal] = None,
        right: typing.Optional[Decimal] = None,
        bottom: typing.Optional[Decimal] = None,
        left: typing.Optional[Decimal] = None,
        zoom: typing.Optional[Decimal] = None,
    ) -> "Document":
        """
        A PDF document may contain a document outline that the conforming reader may display on the screen,
        allowing the user to navigate interactively from one part of the document to another. The outline consists of a
        tree-structured hierarchy of outline items (sometimes called bookmarks), which serve as a visual table of
        contents to display the document’s structure to the user.
        This function adds an outline to this Document
        """
        destination = List().set_is_inline(True)  # type: ignore [attr-defined]
        destination.append(bDecimal(page_nr))
        destination.append(destination_type.value)
        if destination_type == DestinationType.X_Y_Z:
            assert (
                left is not None
                and bottom is None
                and right is None
                and top is not None
                and zoom is not None
            )
            destination.append(bDecimal(left))
            destination.append(bDecimal(top))
            destination.append(bDecimal(zoom))
        if destination_type == DestinationType.FIT:
            assert (
                left is None
                and bottom is None
                and right is None
                and top is None
                and zoom is None
            )
        if destination_type == DestinationType.FIT_H:
            assert (
                left is None
                and bottom is None
                and right is None
                and top is not None
                and zoom is None
            )
            destination.append(bDecimal(top))
        if destination_type == DestinationType.FIT_V:
            assert (
                left is not None
                and bottom is None
                and right is None
                and top is None
                and zoom is None
            )
            destination.append(bDecimal(left))
        if destination_type == DestinationType.FIT_R:
            assert (
                left is not None
                and bottom is not None
                and right is not None
                and top is not None
                and zoom is None
            )
            destination.append(bDecimal(left))
            destination.append(bDecimal(bottom))
            destination.append(bDecimal(right))
            destination.append(bDecimal(top))
        if destination_type == DestinationType.FIT_B_H:
            assert (
                left is None
                and bottom is None
                and right is None
                and top is not None
                and zoom is None
            )
            destination.append(bDecimal(top))
        if destination_type == DestinationType.FIT_B_V:
            assert (
                left is not None
                and bottom is None
                and right is None
                and top is None
                and zoom is None
            )
            destination.append(bDecimal(left))

        # add /Outlines entry in /Root
        if "Outlines" not in self["XRef"]["Trailer"]["Root"]:
            outline_dictionary: Dictionary = Dictionary()
            self["XRef"]["Trailer"]["Root"][Name("Outlines")] = outline_dictionary
            outline_dictionary.set_parent(  # type: ignore [attr-defined]
                self["XRef"]["Trailer"]["Root"][Name("Outlines")]
            )
            outline_dictionary[Name("Type")] = Name("Outlines")
            outline_dictionary[Name("Count")] = bDecimal(0)

        # create entry
        outline = Dictionary()
        outline[Name("Dest")] = destination
        outline[Name("Parent")] = None
        outline[Name("Title")] = String(text)

        # get /Outlines
        outline_dictionary = self["XRef"]["Trailer"]["Root"]["Outlines"]

        # if everything is empty, add the new entry as the only entry
        if "First" not in outline_dictionary or "Last" not in outline_dictionary:
            outline_dictionary[Name("First")] = outline
            outline_dictionary[Name("Last")] = outline
            outline_dictionary[Name("Count")] = bDecimal(1)
            outline[Name("Parent")] = outline_dictionary
            return self

        # helper function to make DFS easier
        def _children(x: Dictionary):
            if "First" not in x:
                return []
            children = [x["First"]]
            while children[-1] != x["Last"]:
                children.append(children[-1]["Next"])
            return children

        # DFS outline(s)
        outlines_done: typing.List[typing.Tuple[int, Dictionary]] = []
        outlines_todo: typing.List[typing.Tuple[int, Dictionary]] = [
            (-1, outline_dictionary)
        ]
        while len(outlines_todo) > 0:
            t = outlines_todo[0]
            outlines_done.append(t)
            outlines_todo.pop(0)
            for c in _children(t[1]):
                outlines_todo.append((t[0] + 1, c))

        # find parent
        parent = [x[1] for x in outlines_done if x[0] == level - 1][-1]

        # update sibling-linking
        if "Last" in parent:
            sibling = parent["Last"]
            sibling[Name("Next")] = outline

        # update parent-linking
        outline[Name("Parent")] = parent
        if "First" not in parent:
            parent[Name("First")] = outline
        if "Count" not in parent:
            parent[Name("Count")] = bDecimal(0)
        parent[Name("Last")] = outline

        # update count
        outline_to_update_count = parent
        while outline_to_update_count:
            outline_to_update_count[Name("Count")] = bDecimal(
                outline_to_update_count["Count"] + Decimal(1)
            )
            if "Parent" in outline_to_update_count:
                outline_to_update_count = outline_to_update_count["Parent"]
            else:
                break

        return self

    #
    # embedded files
    #

    def add_embedded_file(self, file_name: str, file_bytes: bytes) -> "Document":
        """
        If a PDF file contains file specifications that refer to an external file and the PDF file is archived or transmitted,
        some provision should be made to ensure that the external references will remain valid. One way to do this is to
        arrange for copies of the external files to accompany the PDF file. Embedded file streams (PDF 1.3) address
        this problem by allowing the contents of referenced files to be embedded directly within the body of the PDF
        file. This makes the PDF file a self-contained unit that can be stored or transmitted as a single entity. (The
        embedded files are included purely for convenience and need not be directly processed by any conforming reader.)
        This method embeds a file (specified by its name and bytes) into this Document
        """

        # build actual file stream
        stream = Stream()
        stream[Name("Type")] = Name("EmbeddedFile")
        stream[Name("Bytes")] = file_bytes
        stream[Name("Length")] = bDecimal(len(stream[Name("Bytes")]))

        # build /EF NameTree node
        leaf = Dictionary()
        leaf[Name("EF")] = Dictionary()
        leaf[Name("EF")][Name("F")] = stream
        leaf[Name("F")] = String(file_name)
        leaf[Name("Type")] = Name("Filespec")

        # put
        NameTree(self, name=Name("EmbeddedFiles")).put(file_name, leaf)

        # return
        return self

    def get_embedded_files(self) -> typing.Dict[str, bytes]:
        """
        If a PDF file contains file specifications that refer to an external file and the PDF file is archived or transmitted,
        some provision should be made to ensure that the external references will remain valid. One way to do this is to
        arrange for copies of the external files to accompany the PDF file. Embedded file streams (PDF 1.3) address
        this problem by allowing the contents of referenced files to be embedded directly within the body of the PDF
        file. This makes the PDF file a self-contained unit that can be stored or transmitted as a single entity. (The
        embedded files are included purely for convenience and need not be directly processed by any conforming reader.)
        This method returns all embedded files, as a dictionary of names unto bytes
        """
        return {
            str(k): v["EF"]["F"]["DecodedBytes"]
            for k, v in NameTree(self, Name("EmbeddedFiles")).items()
        }

    def get_embedded_file(self, file_name: str) -> typing.Optional[bytes]:
        """
        If a PDF file contains file specifications that refer to an external file and the PDF file is archived or transmitted,
        some provision should be made to ensure that the external references will remain valid. One way to do this is to
        arrange for copies of the external files to accompany the PDF file. Embedded file streams (PDF 1.3) address
        this problem by allowing the contents of referenced files to be embedded directly within the body of the PDF
        file. This makes the PDF file a self-contained unit that can be stored or transmitted as a single entity. (The
        embedded files are included purely for convenience and need not be directly processed by any conforming reader.)
        This method returns the embedded file specified by the given file_name
        """

        for k, v in NameTree(self, Name("EmbeddedFiles")).items():
            if k == file_name:
                return v["EF"]["F"]["DecodedBytes"]

        # default
        return None

    #
    # embedded javascript
    #

    def add_embedded_javascript(self, javascript: str) -> "Document":
        """
        This function appends embedded JavaScript to this Document, returning self.
        :param javascript:  the Javascript str to be appended to this Document
        :return:            self
        """

        # build actual javascript stream
        stream = Stream()
        stream[Name("Type")] = Name("JavaScript")
        stream[Name("DecodedBytes")] = bytes(javascript, "latin1")
        stream[Name("Bytes")] = zlib.compress(stream[Name("DecodedBytes")], 9)
        stream[Name("Length")] = bDecimal(len(stream[Name("Bytes")]))
        stream[Name("Filter")] = Name("FlateDecode")

        # set up NameTree leaf
        leaf = Dictionary()
        leaf[Name("S")] = Name("JavaScript")
        leaf[Name("JS")] = stream

        # put
        nt: NameTree = NameTree(self, name=Name("JavaScript"))
        nt.put("script-{0:03d}.js".format(len(nt)), leaf)

        # return
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for determining which characters (typing.Set[str])
are used by which Font. This is particularly useful when performing Font subsetting.
"""
import typing

from borb.pdf.canvas.event.chunk_of_text_render_event import ChunkOfTextRenderEvent
from borb.pdf.canvas.event.event_listener import EventListener, Event
from borb.pdf.canvas.font.font import Font


class CharacterSetListener(EventListener):
    """
    This implementation of WriteBaseTransformer is responsible for determining which characters (typing.Set[str])
    are used by which Font. This is particularly useful when performing Font subsetting.
    """

    def __init__(self):
        super(CharacterSetListener, self).__init__()
        self._character_set_per_font: typing.Dict[Font, typing.Set[str]] = {}

    def _event_occurred(self, event: Event) -> None:
        if isinstance(event, ChunkOfTextRenderEvent):
            f: Font = event.get_font()
            if f in self._character_set_per_font:
                s: typing.Set[str] = self._character_set_per_font[f]
                for c in event.get_text():
                    s.add(c)
                self._character_set_per_font[f] = s
            else:
                self._character_set_per_font[f] = set([x for x in event.get_text()])

    def get_character_set_per_font(self) -> typing.Dict[Font, typing.Set[str]]:
        """
        This function returns the character set (typing.Set[str]) used by each Font
        :return:    the character set used by each Font
        """
        return self._character_set_per_font



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains the definition of `PageLayout`.
`PageLayout` can be used to add `LayoutElement` objects to a `Page` without
having to specify coordinates.
"""

from borb.pdf.canvas.layout.layout_element import LayoutElement
from borb.pdf.page.page import Page


class PageLayout:
    """
    This class acts as a layoutmanager for `Document` objects.
    Instead of adding `LayoutElements` to a `Page` by calling `layout` (and needing to know the precise coordinates)
    you can simply use this class and its `add` method.
    Any implementation of `PageLayout` should keep track of where `LayoutElement` objects are placed,
    and what the remaining free space is.
    """

    def __init__(self, page: Page):
        self._page = page

    def get_page(self) -> Page:
        """
        This function returns the Page for which this PageLayout is active.
        """
        return self._page

    def add(self, layout_element: LayoutElement) -> "PageLayout":
        """
        This method adds a `LayoutElement` to the current `Page`.
        The specific implementation of `PageLayout` should decide where the `LayoutElement` will be placed.
        """
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This implementation of PageLayout adds left/right/top/bottom margins to a Page
and lays out the content on the Page as if there were was a single column to flow text, images, etc into.
Once this column is full, the next page is automatically created.
"""

import copy
import typing
from decimal import Decimal

from borb.pdf import SingleColumnLayout
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.layout_element import LayoutElement


class SingleColumnLayoutWithOverflow(SingleColumnLayout):
    """
    This implementation of PageLayout adds left/right/top/bottom margins to a Page
    and lays out the content on the Page as if there were was a single column to flow text, images, etc into.
    Once this column is full, the next page is automatically created.
    """

    def __init__(
        self,
        page: "Page",
        horizontal_margin: typing.Optional[Decimal] = None,
        vertical_margin: typing.Optional[Decimal] = None,
    ):
        super(SingleColumnLayoutWithOverflow, self).__init__(
            page, horizontal_margin, vertical_margin
        )

    def add(self, layout_element: LayoutElement) -> "PageLayout":
        """
        This method adds a `LayoutElement` to the current `Page`.
        """

        # anything that isn't a Table gets added as expected
        if layout_element.__class__.__name__ not in [
            "FlexibleColumnWidthTable",
            "FixedColumnWidthTable",
        ]:
            return super(SingleColumnLayout, self).add(layout_element)

        # previous element is used to determine the paragraph spacing
        assert self._page_height is not None
        assert self._page_width is not None
        previous_element_margin_bottom: Decimal = Decimal(0)
        previous_element_y = self._page_height - self._vertical_margin_top
        if self._previous_element is not None:
            previous_element_y = (
                self._previous_element.get_previous_layout_box().get_y()
            )
            previous_element_margin_bottom = self._previous_element.get_margin_bottom()

        # calculate next available height
        available_height: Decimal = (
            previous_element_y
            - self._vertical_margin_bottom
            - self._get_margin_between_elements(self._previous_element, layout_element)
            - max(previous_element_margin_bottom, layout_element.get_margin_top())
            - layout_element.get_margin_bottom()
        )

        # switch to new column if needed
        assert self._page_height
        if available_height < 0:
            self.switch_to_next_column()
            return self.add(layout_element)

        # ask LayoutElement to fit
        lbox: Rectangle = layout_element.get_layout_box(
            Rectangle(
                self._horizontal_margin + layout_element.get_margin_left(),
                Decimal(0),
                self._column_width
                - layout_element.get_margin_right()
                - layout_element.get_margin_left(),
                available_height,
            )
        )
        if lbox.get_height() <= available_height:
            return super(SingleColumnLayout, self).add(layout_element)

        # split
        for t in self._split_table(layout_element, available_height):
            super(SingleColumnLayoutWithOverflow, self).add(t)

        # return
        return self

    def _split_table(
        self, layout_element: LayoutElement, available_height: Decimal
    ) -> typing.List[LayoutElement]:

        # find out at which row we ought to split the Table
        best_row_for_split: typing.Optional[int] = None
        for i in range(0, layout_element._number_of_rows):
            if any([x._row_span != 1 for x in layout_element._get_cells_at_row(i)]):
                continue
            y: Decimal = (
                layout_element._get_cells_at_row(i)[0].get_previous_layout_box().get_y()
            )
            if y < 0:
                continue
            if y < available_height:
                best_row_for_split = i

        # unable to split
        if best_row_for_split is None:
            assert False, (
                "%s is too tall to fit inside column / page."
                % layout_element.__class__.__name__
            )

        # first half of split
        t0 = copy.deepcopy(layout_element)
        t0._number_of_rows = best_row_for_split + 1
        t0._content = [
            x
            for x in t0._content
            if all([y[0] <= best_row_for_split for y in x._table_coordinates])
        ]
        SingleColumnLayoutWithOverflow._prepare_table_for_relayout(t0)

        # second half of split
        t1 = copy.deepcopy(layout_element)
        t1._number_of_rows = layout_element._number_of_rows - best_row_for_split - 1
        t1._content = [
            x
            for x in t1._content
            if all([y[0] > best_row_for_split for y in x._table_coordinates])
        ]
        for tc in t1._content:
            tc._table_coordinates = [
                (y - best_row_for_split - 1, x) for y, x in tc._table_coordinates
            ]
        SingleColumnLayoutWithOverflow._prepare_table_for_relayout(t1)

        # return
        return [t0, t1]

    @staticmethod
    def _prepare_table_for_relayout(layout_element: LayoutElement):
        layout_element._previous_layout_box = None
        layout_element._previous_paint_box = None
        for tc in layout_element._content:
            tc._previous_layout_box = None
            tc._previous_paint_box = None
            tc._forced_layout_box = None
            tc._layout_element._previous_layout_box = None
            tc._layout_element._previous_paint_box = None



# !/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This module contains everything needed to perform low-level tokenization against PDF syntax.
Low-level tokenization aims to separate numbers, strings, names, comments, start of dictionary, start of array, etc
The high-level tokenizer will use this first pass to then build complex objects (streams, dictionaries, etc)
"""
import enum
import io
from typing import Optional


class TokenType(enum.IntEnum):
    """
    This enum represents the various kinds of Token objects the PDF parser can encounter
    """

    NUMBER = 1
    STRING = 2
    HEX_STRING = 3
    NAME = 4
    COMMENT = 5
    START_ARRAY = 6
    END_ARRAY = 7
    START_DICT = 8
    END_DICT = 9
    REF = 10
    OBJ = 11
    END_OBJ = 12
    OTHER = 13
    END_OF_FILE = 14


class Token:
    """
    This class represents a token in PDF syntax.
    A lexical token or simply token is a string with an assigned and thus identified meaning.
    It is structured as a pair consisting of a token name and an optional token value.
    The token name is a category of lexical unit.
    """

    def __init__(self, byte_offset: int, token_type: TokenType, bts: bytes):
        self._byte_offset: int = byte_offset
        self._token_type: TokenType = token_type
        self._bytes: bytes = bts

    def get_byte_offset(self) -> int:
        """
        Get the byte offset of this Token
        """
        return self._byte_offset

    def get_bytes(self) -> bytes:
        """
        Get the bytes of this Token
        """
        return self._bytes

    def get_text(self, encoding: str = "latin1") -> str:
        """
        Get the text of this Token, using a given encoding (default: latin1)
        """
        return self._bytes.decode(encoding)

    def get_token_type(self) -> TokenType:
        """
        Get the TokenType of this Token
        """
        return self._token_type


class LowLevelTokenizer:
    """
    In computer science, lexical analysis, lexing or tokenization is the process of converting a sequence of characters
    (such as in a computer program or web page) into a sequence of tokens (strings with an assigned and thus identified meaning).
    A program that performs lexical analysis may be termed a lexer, tokenizer, or scanner,
    although scanner is also a term for the first stage of a lexer.
    A lexer is generally combined with a parser, which together analyze the syntax of programming languages, web pages,
    and so forth.
    """

    def __init__(self, io_source):
        self._io_source = io_source
        # fmt: off
        self._is_pseudo_digit = set([b'0', b'1', b'2', b'3', b'4', b'5', b'6', b'7', b'8', b'9', b'+', b'-', b'.']).__contains__
        self._is_delimiter = set([b'\x00', b'\t', b'\n', b'\r', b'\x0c', b" ", b'%', b'(', b')', b'/', b'<', b'>', b'[', b']']).__contains__
        self._is_whitespace = set([b'\x00', b'\t', b'\n', b'\r', b'\x0c', b' ']).__contains__
        # fmt: on

    def next_non_comment_token(self) -> Optional[Token]:
        """
        This function retrieves the next non-comment Token.
        It returns None if no such Token exists (end of stream/file)
        """
        t = self.next_token()
        while t is not None and t.get_token_type() == TokenType.COMMENT:
            t = self.next_token()
        return t

    def next_token(self) -> Optional[Token]:
        """
        This function retrieves the next Token.
        It returns None if no such Token exists (end of stream/file)
        """
        ch = self._next_byte()
        if len(ch) == 0:
            return None

        # skip whitespace
        while len(ch) > 0 and self._is_whitespace(ch):
            ch = self._next_byte()

        # START_ARRAY
        if ch == b"[":
            return Token(self._io_source.tell() - 1, TokenType.START_ARRAY, b"[")

        # END ARRAY
        if ch == b"]":
            return Token(self._io_source.tell() - 1, TokenType.END_ARRAY, b"]")

        # NAME
        out_str: bytearray = bytearray()
        if ch == b"/":
            out_str = bytearray(b"/")
            out_pos = self._io_source.tell() - 1
            while True:
                ch = self._next_byte()
                if len(ch) == 0:
                    break
                if self._is_delimiter(ch):
                    break
                out_str += ch
            if len(ch) != 0:
                self._prev_byte()
            return Token(out_pos, TokenType.NAME, bytes(out_str))

        # END_DICT
        if ch == b">":
            out_pos = self._io_source.tell() - 1
            ch = self._next_byte()
            # CHECK UNEXPECTED CHARACTER AFTER FIRST >
            assert ch == b">", "Unexpected character at end of dictionary."
            return Token(out_pos, TokenType.END_DICT, b">>")

        # COMMENT
        if ch == b"%":
            out_str = bytearray([])
            out_pos = self._io_source.tell() - 1
            while len(ch) != 0 and ch != b"\r" and ch != b"\n":
                out_str += ch
                ch = self._next_byte()
            if len(ch) != 0:
                self._prev_byte()
            return Token(out_pos, TokenType.COMMENT, bytes(out_str))

        # HEX_STRING OR DICT
        if ch == b"<":
            out_pos = self._io_source.tell() - 1
            ch = self._next_byte()

            # DICT
            if ch == b"<":
                return Token(out_pos, TokenType.START_DICT, b"<<")

            # empty hex string
            if ch == b">":
                return Token(out_pos, TokenType.HEX_STRING, b"<>")

            # HEX_STRING
            out_str = bytearray(b"<")
            out_str += ch
            while True:
                ch = self._next_byte()
                if len(ch) == 0:
                    break
                out_str += ch
                if ch == b">":
                    break
            return Token(out_pos, TokenType.HEX_STRING, bytes(out_str))

        # NUMBER
        if self._is_pseudo_digit(ch):
            out_str = bytearray([])
            out_pos = self._io_source.tell() - 1
            while len(ch) != 0 and self._is_pseudo_digit(ch):
                out_str += ch
                ch = self._next_byte()
            if len(ch) != 0:
                self._prev_byte()
            return Token(out_pos, TokenType.NUMBER, bytes(out_str))

        # STRING
        if ch == b"(":
            bracket_nesting_level = 1
            out_str = bytearray(b"(")
            out_pos = self._io_source.tell() - 1
            while True:
                ch = self._next_byte()
                if len(ch) == 0:
                    break
                # escape char
                if ch == b"\\":
                    ch = self._next_byte()
                    out_str += b"\\"
                    out_str += ch
                    continue
                if ch == b"(":
                    bracket_nesting_level += 1
                if ch == b")":
                    bracket_nesting_level -= 1
                out_str += ch
                if bracket_nesting_level == 0:
                    break
            assert len(ch) != 0
            assert out_str[-1] != b"\\"
            return Token(out_pos, TokenType.STRING, bytes(out_str))

        # OTHER
        out_str = bytearray([])
        out_pos = self._io_source.tell() - 1
        while len(ch) != 0 and not self._is_delimiter(ch):
            out_str += ch
            ch = self._next_byte()
        if len(ch) != 0:
            self._prev_byte()
        return Token(out_pos, TokenType.OTHER, bytes(out_str))

    def seek(self, pos: int, whence: int = io.SEEK_SET):
        """
        Change the stream position to the given byte offset. offset is interpreted relative to the position indicated by whence.
        The default value for whence is SEEK_SET. Values for whence are:
        SEEK_SET or 0 – start of the stream (the default); offset should be zero or positive
        SEEK_CUR or 1 – current stream position; offset may be negative
        SEEK_END or 2 – end of the stream; offset is usually negative
        Return the new absolute position.
        """
        return self._io_source.seek(pos, whence)

    def tell(self) -> int:
        """
        Return the current stream position.
        """
        return self._io_source.tell()

    def _next_byte(self):
        return self._io_source.read(1)

    def _prev_byte(self):
        return self._io_source.seek(-1, io.SEEK_CUR)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Type 3 fonts differ from the other fonts supported by PDF. A Type 3 font dictionary defines the font; font
dictionaries for other fonts simply contain information about the font and refer to a separate font program for the
actual glyph descriptions. In Type 3 fonts, glyphs shall be defined by streams of PDF graphics operators. These
streams shall be associated with glyph names. A separate encoding entry shall map character codes to the
appropriate glyph names for the glyphs.
"""

import logging
import typing

from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Name
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.font.simple_font.font_type_1 import Type1Font

logger = logging.getLogger(__name__)


class Type3Font(Type1Font):
    """
    Type 3 fonts differ from the other fonts supported by PDF. A Type 3 font dictionary defines the font; font
    dictionaries for other fonts simply contain information about the font and refer to a separate font program for the
    actual glyph descriptions. In Type 3 fonts, glyphs shall be defined by streams of PDF graphics operators. These
    streams shall be associated with glyph names. A separate encoding entry shall map character codes to the
    appropriate glyph names for the glyphs.
    """

    def __init__(self):
        super(Type3Font, self).__init__()
        self[Name("Subtype")] = Name("Type3")

    def get_ascent(self) -> bDecimal:
        """
        This function returns the maximum height above the baseline reached by glyphs in this font.
        The height of glyphs for accented characters shall be excluded.
        """
        if "FontDescriptor" in self and "Ascent" in "FontDescriptor":
            return self["FontDescriptor"]["Ascent"]
        logger.debug(
            "Type3Font does not have an `Ascent` entry in its `FontDescriptor` dictionary."
        )
        return bDecimal(0)  # TODO

    def get_descent(self) -> bDecimal:
        """
        This function returns the maximum depth below the baseline reached by glyphs in this font.
        The value shall be a negative number.
        """
        if "FontDescriptor" in self and "Descent" in "FontDescriptor":
            return self["FontDescriptor"]["Descent"]
        logger.debug(
            "Type3Font does not have an `Descent` entry in its `FontDescriptor` dictionary."
        )
        return bDecimal(0)  # TODO

    def _empty_copy(self) -> "Font":
        return Type3Font()

    def __deepcopy__(self, memodict={}):
        # fmt: off
        f_out: Font = super(Type3Font, self).__deepcopy__(memodict)
        f_out[Name("Subtype")] = Name("Type3")
        f_out._character_identifier_to_unicode_lookup: typing.Dict[int, str] = {k: v for k, v in self._character_identifier_to_unicode_lookup.items()}
        f_out._unicode_lookup_to_character_identifier: typing.Dict[str, int] = {k: v for k, v in self._unicode_lookup_to_character_identifier.items()}
        return f_out
        # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Same as RG but used for nonstroking operations.
"""
import typing
from decimal import Decimal

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.color.color import RGBColor
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetRGBNonStroking(CanvasOperator):
    """
    Same as RG but used for nonstroking operations.
    """

    def __init__(self):
        super().__init__("rg", 3)

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the rg operator
        """
        # fmt: off
        assert isinstance(operands[0], Decimal), "operand 0 of rg operator must be of type Decimal"
        assert isinstance(operands[1], Decimal), "operand 1 of rg operator must be of type Decimal"
        assert isinstance(operands[2], Decimal), "operand 2 of rg operator must be of type Decimal"
        # fmt: on

        # set non_stroke_color
        canvas = canvas_stream_processor.get_canvas()
        canvas.graphics_state.non_stroke_color = RGBColor(
            operands[0], operands[1], operands[2]
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
There are several types of simple fonts, all of which have these properties:

•   Glyphs in the font shall be selected by single-byte character codes obtained from a string that is shown by
    the text-showing operators. Logically, these codes index into a table of 256 glyphs; the mapping from
    codes to glyphs is called the font’s encoding. Under some circumstances, the encoding may be altered by
    means described in 9.6.6, "Character Encoding".

•   Each glyph shall have a single set of metrics, including a horizontal displacement or width, as described in
    9.2.4, "Glyph Positioning and Metrics"; that is, simple fonts support only horizontal writing mode.

•   Except for Type 0 fonts, Type 3 fonts in non-Tagged PDF documents, and certain standard Type 1 fonts,
    every font dictionary shall contain a subsidiary dictionary, the font descriptor, containing font-wide metrics
    and other attributes of the font; see 9.8, "Font Descriptors". Among those attributes is an optional font file
    stream containing the font program.
"""
from borb.pdf.canvas.font.font import Font


class SimpleFont(Font):
    """
    There are several types of simple fonts, all of which have these properties:

    •   Glyphs in the font shall be selected by single-byte character codes obtained from a string that is shown by
        the text-showing operators. Logically, these codes index into a table of 256 glyphs; the mapping from
        codes to glyphs is called the font’s encoding. Under some circumstances, the encoding may be altered by
        means described in 9.6.6, "Character Encoding".

    •   Each glyph shall have a single set of metrics, including a horizontal displacement or width, as described in
        9.2.4, "Glyph Positioning and Metrics"; that is, simple fonts support only horizontal writing mode.

    •   Except for Type 0 fonts, Type 3 fonts in non-Tagged PDF documents, and certain standard Type 1 fonts,
        every font dictionary shall contain a subsidiary dictionary, the font descriptor, containing font-wide metrics
        and other attributes of the font; see 9.8, "Font Descriptors". Among those attributes is an optional font file
        stream containing the font program.
    """

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
The purpose of a line annotation (PDF 1.3) is to display a single straight line on the page. When opened, it shall
display a pop-up window containing the text of the associated note. Table 175 shows the annotation dictionary
entries specific to this type of annotation.
"""
import typing
from decimal import Decimal

from borb.io.read.types import Name, List, Decimal as bDecimal
from borb.pdf.canvas.color.color import HexColor, Color
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.annotation import Annotation
from borb.pdf.canvas.layout.annotation.polyline_annotation import LineEndStyleType


class LineAnnotation(Annotation):
    """
    The purpose of a line annotation (PDF 1.3) is to display a single straight line on the page. When opened, it shall
    display a pop-up window containing the text of the associated note. Table 175 shows the annotation dictionary
    entries specific to this type of annotation.
    """

    def __init__(
        self,
        start_point: typing.Tuple[Decimal, Decimal],
        end_point: typing.Tuple[Decimal, Decimal],
        left_line_end_style: LineEndStyleType = LineEndStyleType.NONE,
        right_line_end_style: LineEndStyleType = LineEndStyleType.NONE,
        stroke_color: Color = HexColor("000000"),
    ):

        x = min([start_point[0], end_point[0]])
        y = min([start_point[1], end_point[1]])
        w = max([start_point[0], end_point[0]]) - x
        h = max([start_point[1], end_point[1]]) - y

        # create generic annotation
        super(LineAnnotation, self).__init__(
            bounding_box=Rectangle(x, y, w, h), color=stroke_color
        )

        # (Required) The type of annotation that this dictionary describes; shall be
        # Line for a line annotation.
        self[Name("Subtype")] = Name("Line")

        # (Required) An array of four numbers, [ x 1 y 1 x 2 y 2 ], specifying the
        # starting and ending coordinates of the line in default user space.
        # If the LL entry is present, this value shall represent the endpoints of the
        # leader lines rather than the endpoints of the line itself; see Figure 60.
        self[Name("L")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self["L"].append(start_point[0])
        self["L"].append(start_point[1])
        self["L"].append(end_point[0])
        self["L"].append(end_point[1])

        # (Optional; PDF 1.4) An array of two names specifying the line ending
        # styles that shall be used in drawing the line. The first and second
        # elements of the array shall specify the line ending styles for the endpoints
        # defined, respectively, by the first and second pairs of coordinates, (x 1 , y 1 )
        # and (x 2 , y 2 ), in the L array. Table 176 shows the possible values. Default
        # value: [ /None /None ].
        self[Name("LE")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self["LE"].append(left_line_end_style.value)
        self["LE"].append(right_line_end_style)

        # (Optional; PDF 1.4) An array of numbers that shall be in the range 0.0 to
        # 1.0 and shall specify the interior color with which to fill the annotation’s
        # rectangle or ellipse. The number of array elements determines the colour
        # space in which the colour shall be defined
        if stroke_color is not None:
            self[Name("IC")] = List().set_is_inline(True)  # type: ignore [attr-defined]
            self["IC"].append(bDecimal(stroke_color.to_rgb().red))
            self["IC"].append(bDecimal(stroke_color.to_rgb().green))
            self["IC"].append(bDecimal(stroke_color.to_rgb().blue))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
An annotation associates an object such as a note, sound, or movie with a location on a page of a PDF
document, or provides a way to interact with the user by means of the mouse and keyboard. PDF includes a
wide variety of standard annotation types, described in detail in 12.5.6, “Annotation Types.”
"""
import datetime
import typing
from decimal import Decimal

from borb.io.read.types import Dictionary, Name, List, Decimal as bDecimal, String
from borb.pdf.canvas.color.color import Color
from borb.pdf.canvas.geometry.rectangle import Rectangle


class Annotation(Dictionary):
    """
    An annotation associates an object such as a note, sound, or movie with a location on a page of a PDF
    document, or provides a way to interact with the user by means of the mouse and keyboard. PDF includes a
    wide variety of standard annotation types, described in detail in 12.5.6, “Annotation Types.”
    """

    @staticmethod
    def _timestamp_to_str() -> str:
        timestamp_str = "D:"
        now = datetime.datetime.now()
        for n in [now.year, now.month, now.day, now.hour, now.minute, now.second]:
            timestamp_str += "{0:02}".format(n)
        timestamp_str += "Z00"
        return timestamp_str

    def __init__(
        self,
        bounding_box: Rectangle,
        contents: typing.Optional[str] = None,
        color: typing.Optional[Color] = None,
        horizontal_corner_radius: typing.Optional[Decimal] = None,
        vertical_corner_radius: typing.Optional[Decimal] = None,
        border_width: typing.Optional[Decimal] = None,
    ):
        super(Annotation, self).__init__()

        # (Optional) The type of PDF object that this dictionary describes; if
        # present, shall be Annot for an annotation dictionary.
        self[Name("Type")] = Name("Annot")

        # (Required) The annotation rectangle, defining the location of the
        # annotation on the page in default user space units.
        self[Name("Rect")] = List().set_is_inline(True)  # type: ignore [attr-defined]
        self["Rect"].append(bDecimal(bounding_box.get_x()))
        self["Rect"].append(bDecimal(bounding_box.get_y()))
        self["Rect"].append(bDecimal(bounding_box.get_x() + bounding_box.get_width()))
        self["Rect"].append(bDecimal(bounding_box.get_y() + bounding_box.get_height()))

        # (Optional) Text that shall be displayed for the annotation or, if this type of
        # annotation does not display text, an alternate description of the
        # annotation’s contents in human-readable form. In either case, this text is
        # useful when extracting the document’s contents in support of
        # accessibility to users with disabilities or for other purposes (see 14.9.3,
        # “Alternate Descriptions”). See 12.5.6, “Annotation Types” for more
        # details on the meaning of this entry for each annotation type.
        if contents is not None:
            self[Name("Contents")] = String(contents)

        # (Optional; PDF 1.4) The annotation name, a text string uniquely
        # identifying it among all the annotations on its page.
        len_annots: int = 0
        self[Name("NM")] = String("annotation-{0:03d}".format(len_annots))

        # (Optional; PDF 1.1) The date and time when the annotation was most
        # recently modified. The format should be a date string as described in
        # 7.9.4, “Dates,” but conforming readers shall accept and display a string
        # in any format.
        self[Name("M")] = String(self._timestamp_to_str())

        # (Optional; PDF 1.1) A set of flags specifying various characteristics of
        # the annotation (see 12.5.3, “Annotation Flags”). Default value: 0.
        self[Name("F")] = bDecimal(4)

        # (Optional; PDF 1.2) An appearance dictionary specifying how the
        # annotation shall be presented visually on the page (see 12.5.5,
        # “Appearance Streams”). Individual annotation handlers may ignore this
        # entry and provide their own appearances.
        # self[Name("AP")] = None

        # (Required if the appearance dictionary AP contains one or more
        # subdictionaries; PDF 1.2) The annotation’s appearance state, which
        # selects the applicable appearance stream from an appearance
        # subdictionary (see Section 12.5.5, “Appearance Streams”).
        # self[Name("AS")] = None

        # Optional) An array specifying the characteristics of the annotation’s
        # border, which shall be drawn as a rounded rectangle.
        # (PDF 1.0) The array consists of three numbers defining the horizontal
        # corner radius, vertical corner radius, and border width, all in default user
        # space units. If the corner radii are 0, the border has square (not rounded)
        # corners; if the border width is 0, no border is drawn.
        # (PDF 1.1) The array may have a fourth element, an optional dash array
        # defining a pattern of dashes and gaps that shall be used in drawing the
        # border. The dash array shall be specified in the same format as in the
        # line dash pattern parameter of the graphics state (see 8.4.3.6, “Line
        # Dash Pattern”).
        if (
            horizontal_corner_radius is not None
            and vertical_corner_radius is not None
            and border_width is not None
        ):
            self[Name("Border")] = List().set_is_inline(True)  # type: ignore [attr-defined]
            self["Border"].append(bDecimal(horizontal_corner_radius))
            self["Border"].append(bDecimal(vertical_corner_radius))
            self["Border"].append(bDecimal(border_width))

        # (Optional; PDF 1.1) An array of numbers in the range 0.0 to 1.0,
        # representing a colour used for the following purposes:
        # The background of the annotation’s icon when closed
        # The title bar of the annotation’s pop-up window
        # The border of a link annotation
        # The number of array elements determines the colour space in which the
        # colour shall be defined
        if color is not None:
            self[Name("C")] = List().set_is_inline(True)  # type: ignore [attr-defined]
            self["C"].append(bDecimal(color.to_rgb().red))
            self["C"].append(bDecimal(color.to_rgb().green))
            self["C"].append(bDecimal(color.to_rgb().blue))

        # (Required if the annotation is a structural content item; PDF 1.3) The
        # integer key of the annotation’s entry in the structural parent tree (see
        # 14.7.4.4, “Finding Structure Elements from Content Items”)
        # self[Name("StructParent")] = None

        # (Optional; PDF 1.5) An optional content group or optional content
        # membership dictionary (see 8.11, “Optional Content”) specifying the
        # optional content properties for the annotation. Before the annotation is
        # drawn, its visibility shall be determined based on this entry as well as the
        # annotation flags specified in the F entry (see 12.5.3, “Annotation Flags”).
        # If it is determined to be invisible, the annotation shall be skipped, as if it
        # were not in the document.
        # self[Name("OC")] = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible
for writing Dictionary objects of /Type /Page
"""
import logging
from typing import Optional

from borb.io.read.types import AnyPDFType, Dictionary, Name
from borb.io.write.font.subsetter import Subsetter
from borb.io.write.object.dictionary_transformer import DictionaryTransformer
from borb.io.write.transformer import WriteTransformerState
from borb.pdf.document.document import Document

logger = logging.getLogger(__name__)


class PageTransformer(DictionaryTransformer):
    """
    This implementation of WriteBaseTransformer is responsible
    for writing Dictionary objects of /Type /Page
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be converted represents an /Page Dictionary
        """
        return isinstance(any, Dictionary) and "Type" in any and any["Type"] == "Page"

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes a /Page Dictionary to a byte stream
        """
        # fmt: off
        assert isinstance(object_to_transform, Dictionary)
        assert (context is not None), "context must be defined in order to write Page objects."
        assert context.root_object is not None, "context.root_object must be defined in order to write Page objects."
        assert isinstance(context.root_object, Document), "context.root_object must be of type Document in order to write Page objects."
        # fmt: on

        pages_dict = context.root_object["XRef"]["Trailer"]["Root"]["Pages"]

        # add /Parent reference to /Pages
        object_to_transform[Name("Parent")] = self.get_reference(pages_dict, context)

        # mark some keys as non-referencable
        for k in ["ArtBox", "BleedBox", "CropBox", "MediaBox", "TrimBox"]:
            if k in object_to_transform:
                object_to_transform[k].set_is_inline(True)  # type: ignore [attr-defined]

        # apply subsetting
        if context.apply_font_subsetting:
            page = Subsetter.apply(object_to_transform)

        # delegate to super
        super(PageTransformer, self).transform(object_to_transform, context)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
In metal typesetting, a font was a particular size, weight and style of a typeface.
Each font was a matched set of type, one piece (called a "sort") for each glyph, and a typeface consisting of a range of fonts that shared an overall design.

In modern usage, with the advent of digital typography, "font" is frequently synonymous with "typeface".
Each style is in a separate "font file"—for instance, the typeface "Bulmer" may include the fonts "Bulmer roman",
"Bulmer", "Bulmer bold" and "Bulmer extended"—but the term "font" might be applied either to one of these alone or to the whole typeface.

In both traditional typesetting and modern usage, the word "font" refers to the delivery mechanism of the typeface design.
In traditional typesetting, the font would be made from metal or wood.
Today, the font is a digital file.
"""
import copy
import io
import typing
from decimal import Decimal

from borb.io.read.tokenize.high_level_tokenizer import HighLevelTokenizer
from borb.io.read.tokenize.low_level_tokenizer import Token, TokenType
from borb.io.read.types import Decimal as bDecimal
from borb.io.read.types import Dictionary, List, Name


class Font(Dictionary):
    """
    In both traditional typesetting and modern usage, the word "font" refers to the delivery mechanism of the typeface design.
    In traditional typesetting, the font would be made from metal or wood.
    Today, the font is a digital file.
    """

    def character_identifier_to_unicode(
        self, character_identifier: int
    ) -> typing.Optional[str]:
        """
        This function maps a character identifier to its unicode str.
        If no such mapping exists, this function returns None.
        """
        return None

    def unicode_to_character_identifier(self, unicode: str) -> typing.Optional[int]:
        """
        This function maps a unicode str to its character identifier.
        If no such mapping exists, this function returns None.
        """
        return None

    def get_width(self, character_identifier: int) -> typing.Optional[bDecimal]:
        """
        This function returns the width (in text space) of a given character identifier.
        If this Font is unable to represent the glyph that corresponds to the character identifier,
        this function returns None
        """
        return None

    def get_ascent(self) -> bDecimal:
        """
        This function returns the maximum height above the baseline reached by glyphs in this font.
        The height of glyphs for accented characters shall be excluded.
        """
        return bDecimal(0)

    def get_descent(self) -> bDecimal:
        """
        This function returns the maximum depth below the baseline reached by glyphs in this font.
        The value shall be a negative number.
        """
        return bDecimal(0)

    def get_font_name(self) -> typing.Optional[str]:
        """
        This function returns the name of the Font
        e.g. "Helvetica"
        """
        if "BaseFont" in self:
            return str(self["BaseFont"])
        if "FontFamily" in self:
            return str(self["FontFamily"])
        return None

    def get_space_character_width_estimate(self) -> Decimal:
        """
        This function estimates the width of the space character (unicode 32) in this Font.
        If the Font contains the character, this Font will return the corresponding width.

        If the Font does not contain the character, its width may be derived from the
        MissingWidth entry in the FontDescriptor, or the AvgWidth entry.

        If the Font is a composite Font, the DW entry of its DescendantFont is used.

        If all previously mentioned approaches fail, the width is estimated based on characters
        that may be present in the Font. (e.g. the width of 'A' is typically twice that of ' ').
        """
        # 1. if space is defined, and the width of space is defined, return that
        character_identifier: typing.Optional[
            int
        ] = self.unicode_to_character_identifier(" ")
        width: typing.Optional[Decimal] = None
        if character_identifier is not None:
            width = self.get_width(character_identifier)
            if width is not None:
                return width
        # 2. MissingWidth
        if "FontDescriptor" in self and "MissingWidth" in self["FontDescriptor"]:
            return self["FontDescriptor"]["MissingWidth"] / Decimal(2)
        # 3. AvgWidth
        if "FontDescriptor" in self and "AvgWidth" in self["FontDescriptor"]:
            return self["FontDescriptor"]["AvgWidth"] / Decimal(2)
        # 3. default width
        if (
            "DescendantFonts" in self
            and isinstance(self["DescendantFonts"], List)
            and len(self["DescendantFonts"]) == 1
            and "DW" in self["DescendantFonts"][0]
        ):
            return self["DescendantFonts"][0]["DW"] / Decimal(2)
        # 4. other characters may be defined, which give us a clue
        # fmt: off
        char_to_space_width_ratio: typing.Dict[str, Decimal] = {
            "a": Decimal("0.500"), "b": Decimal("0.500"), "c": Decimal("0.556"),
            "d": Decimal("0.500"), "e": Decimal("0.500"), "f": Decimal("1.000"),
            "g": Decimal("0.500"), "h": Decimal("0.500"), "i": Decimal("1.252"),
            "j": Decimal("1.252"), "k": Decimal("0.556"), "l": Decimal("1.252"),
            "m": Decimal("0.334"), "n": Decimal("0.500"), "o": Decimal("0.500"),
            "p": Decimal("0.500"), "q": Decimal("0.500"), "r": Decimal("0.835"),
            "s": Decimal("0.556"), "t": Decimal("1.000"), "u": Decimal("0.500"),
            "v": Decimal("0.556"), "w": Decimal("0.385"), "x": Decimal("0.556"),
            "y": Decimal("0.556"), "z": Decimal("0.556"), "0": Decimal("0.500"),
            "1": Decimal("0.500"), "2": Decimal("0.500"), "3": Decimal("0.500"),
            "4": Decimal("0.500"), "5": Decimal("0.500"), "6": Decimal("0.500"),
            "7": Decimal("0.500"), "8": Decimal("0.500"), "9": Decimal("0.500"),
            "A": Decimal("0.417"), "B": Decimal("0.417"), "C": Decimal("0.385"),
            "D": Decimal("0.385"), "E": Decimal("0.417"), "F": Decimal("0.455"),
            "G": Decimal("0.357"), "H": Decimal("0.385"), "I": Decimal("1.000"),
            "J": Decimal("0.556"), "K": Decimal("0.417"), "L": Decimal("0.500"),
            "M": Decimal("0.334"), "N": Decimal("0.385"), "O": Decimal("0.357"),
            "P": Decimal("0.417"), "Q": Decimal("0.357"), "R": Decimal("0.385"),
            "S": Decimal("0.417"), "T": Decimal("0.455"), "U": Decimal("0.385"),
            "V": Decimal("0.417"), "W": Decimal("0.294"), "X": Decimal("0.417"),
            "Y": Decimal("0.417"), "Z": Decimal("0.455"),
        }
        # fmt: on
        for k, v in char_to_space_width_ratio.items():
            character_identifier = self.unicode_to_character_identifier(k)
            if character_identifier is not None:
                width = self.get_width(character_identifier)
                if width is not None:
                    return bDecimal(width * v)
        # 5. helvetica
        return Decimal(278)

    # fmt: off
    @staticmethod
    def _read_cmap(cmap_bytes: bytes) -> typing.Dict[int, str]:

        out_map: typing.Dict[int, str] = {}
        cmap_tokenizer: HighLevelTokenizer = HighLevelTokenizer(io.BytesIO(cmap_bytes))

        # process stream
        prev_token: typing.Optional[Token] = None
        number_of_bytes = len(cmap_bytes)
        while cmap_tokenizer.tell() < number_of_bytes:
            token: typing.Optional[Token] = cmap_tokenizer.next_non_comment_token()
            assert token is not None

            # beginbfchar
            if token.get_text() == "beginbfchar":
                assert prev_token is not None
                number_of_lines_001: int = int(prev_token.get_text())
                for _ in range(0, number_of_lines_001):
                    token = cmap_tokenizer.next_non_comment_token()
                    assert token is not None
                    char_code: int = int(token.get_text()[1:-1], 16)

                    token = cmap_tokenizer.next_non_comment_token()
                    assert token is not None
                    if token.get_text().startswith("<") and token.get_text().endswith(">"):
                        unicode_str: str = str(token.get_text())[1:-1].replace(" ","")
                        unicode_str = "".join([chr(int(unicode_str[j: j + 4], 16)) for j in range(0, len(unicode_str), 4)])
                    elif token.get_text().startswith("/"):
                        # TODO
                        assert False, "Unsupported CMAP syntax"
                    else:
                        assert False, "Invalid CMAP"
                    out_map[char_code] = unicode_str

            if token.get_text() == "begincidrange":
                assert prev_token is not None
                number_of_lines_002: int = int(prev_token.get_text())
                for _ in range(0, number_of_lines_002):

                    token = cmap_tokenizer.next_non_comment_token()
                    assert token is not None
                    char_code_start_002: int = int(token.get_text()[1:-1], 16)

                    token = cmap_tokenizer.next_non_comment_token()
                    assert token is not None
                    char_code_stop_002: int = int(token.get_text()[1:-1], 16)

                    token = cmap_tokenizer.next_non_comment_token()
                    assert token is not None

                    # <FFFF> <FFFF> 0000
                    if token.get_token_type() == TokenType.NUMBER:
                        unicode_base: int = int(token.get_text())
                        for i in range(char_code_start_002, char_code_stop_002 + 1):
                            out_map[i] = chr(unicode_base + (i - char_code_start_002))

            # beginbfrange
            if token.get_text() == "beginbfrange":
                assert prev_token is not None
                number_of_lines_003: int = int(prev_token.get_text())
                for _ in range(0, number_of_lines_003):

                    token = cmap_tokenizer.next_non_comment_token()
                    assert token is not None
                    char_code_start_003: int = int(token.get_text()[1:-1], 16)

                    token = cmap_tokenizer.next_non_comment_token()
                    assert token is not None
                    char_code_stop_003: int = int(token.get_text()[1:-1], 16)

                    token = cmap_tokenizer.next_non_comment_token()
                    assert token is not None

                    # <FFFF> <FFFF> <FFFF>
                    if token.get_token_type() == TokenType.HEX_STRING:
                        unicode_base_str: str = str(token.get_text())[1:-1]
                        for i in range(char_code_start_003, char_code_stop_003 + 1):
                            unicode_hex: str = hex(int(unicode_base_str, 16) + (i - char_code_start_003))[2:]
                            unicode_hex = ("".join(["0" for _ in range(0, 4 - len(unicode_hex) % 4)]) + unicode_hex)
                            unicode_hex = "".join([chr(int(unicode_hex[j: j + 4], 16)) for j in range(0, len(unicode_hex), 4)])
                            out_map[i] = unicode_hex
                        continue

                    # <FFFF> <FFFF> [ <FFFF>* ]
                    if token.get_token_type() == TokenType.START_ARRAY:
                        for i in range(char_code_start_003, char_code_stop_003 + 1):
                            token = cmap_tokenizer.next_non_comment_token()
                            assert token is not None
                            unicode_base_str_003: str = str(token.get_text())[1:-1]
                            unicode_hex = ("".join(["0" for _ in range(0, 4 - len(unicode_base_str_003) % 4)]) + unicode_base_str_003)
                            unicode_hex = "".join([chr(int(unicode_hex[j: j + 4], 16)) for j in range(0, len(unicode_hex), 4)])
                            out_map[i] = unicode_hex
                        # read END_ARRAY
                        token = cmap_tokenizer.next_non_comment_token()
                        assert token is not None
                        assert token.get_token_type() == TokenType.END_ARRAY
                        continue

            # set previous token
            prev_token = token

        # return
        return out_map
    # fmt: on

    def _empty_copy(self) -> "Font":
        assert False

    def __deepcopy__(self, memodict={}) -> "Font":
        out: Font = self._empty_copy()
        # Type
        out[Name("Type")] = Name("Font")
        # BaseFont
        if "BaseFont" in self:
            out[Name("BaseFont")] = Name(str(self["BaseFont"]))
        # FirstChar
        if "FirstChar" in self:
            out[Name("FirstChar")] = self["FirstChar"]
        # LastChar
        if "LastChar" in self:
            out[Name("LastChar")] = self["LastChar"]
        # Widths
        if "Widths" in self:
            out[Name("Widths")] = List()
            for k in self["Widths"]:
                out[Name("Widths")].append(k)
        # FontDescriptor
        if "FontDescriptor" in self:
            out[Name("FontDescriptor")] = self._copy_font_descriptor(
                self["FontDescriptor"]
            )
        # Encoding
        if "Encoding" in self:
            # Name
            if isinstance(self["Encoding"], Name):
                out[Name("Encoding")] = Name(str(self["Encoding"]))
            # Dictionary
            if isinstance(self["Encoding"], Dictionary):
                out[Name("Encoding")] = Dictionary()
                out["Encoding"][Name("Type")] = Name("Encoding")
                if "BaseEncoding" in self["Encoding"]:
                    out["Encoding"][Name("BaseEncoding")] = Name(
                        str(self["Encoding"]["BaseEncoding"])
                    )
                if "Differences" in self["Encoding"]:
                    l = List()
                    for x in self["Encoding"]["Differences"]:
                        l.append(x)
                    out["Encoding"][Name("Differences")] = l
        # ToUnicode
        if "ToUnicode" in self:
            out[Name("ToUnicode")] = copy.deepcopy(self["ToUnicode"])
        # FontBBox
        if "FontBBox" in self:
            out[Name("FontBBox")] = List()
            for x in self["FontBBox"]:
                out["FontBBox"].append(x)
        # FontMatrix
        if "FontMatrix" in self:
            out[Name("FontMatrix")] = List()
            for x in self["FontMatrix"]:
                out["FontMatrix"].append(x)
        # CharProcs
        # Resources
        # CIDSystemInfo
        if "CIDSystemInfo" in self:
            out[Name("CIDSystemInfo")] = Dictionary()
            out["CIDSystemInfo"][Name("Registry")] = self["CIDSystemInfo"]["Registry"]
            out["CIDSystemInfo"][Name("Ordering")] = self["CIDSystemInfo"]["Ordering"]
            out["CIDSystemInfo"][Name("Supplement")] = self["CIDSystemInfo"][
                "Supplement"
            ]
        # DW
        if "DW" in self:
            out[Name("DW")] = self["DW"]
        # W
        if "W" in self:
            out[Name("W")] = List()
            for x in self["W"]:
                if isinstance(x, bDecimal):
                    out["W"].append(x)
                if isinstance(x, List):
                    l = List()
                    for y in x:
                        l.append(y)
                    out["W"].append(l)
        # DescendantFonts
        if "DescendantFonts" in self:
            out[Name("DescendantFonts")] = List()
            out["DescendantFonts"].append(
                self["DescendantFonts"][0].__deepcopy__(memodict)
            )
        # DW2
        if "DW2" in self:
            out[Name("DW2")] = List()
            for x in self["DW2"]:
                out["DW2"].append(x)
        # W2
        # CIDToGIDMap
        # default
        for k, v in self.items():
            if k not in out:
                out[k] = copy.deepcopy(v, memodict)
        return out

    def _copy_font_descriptor(self, font_descriptor_to_copy: Dictionary) -> Dictionary:
        f0: Dictionary = font_descriptor_to_copy
        f1: Dictionary = self["FontDescriptor"]
        f1[Name("Type")] = f0.get("Type", Name("FontDescriptor"))
        f1[Name("FontName")] = f0["FontName"]
        if "FontFamily" in f0:
            f1[Name("FontFamily")] = f0["FontFamily"]
        if "FontStretch" in f0:
            f1[Name("FontStretch")] = f0["FontStretch"]
        if "FontWeight" in f0:
            f1[Name("FontWeight")] = f0["FontWeight"]
        f1[Name("Flags")] = f0["Flags"]
        if "FontBBox" in f0 and False:  # TODO
            f1[Name("FontBBox")] = List().set_is_inline(True)  # type: ignore [attr-defined]
            for i in range(0, len(f0["FontBBox"])):
                f1["FontBBox"].append(f0["FontBBox"][i])
        f1[Name("ItalicAngle")] = f0["ItalicAngle"]
        if "Ascent" in f0:
            f1[Name("Ascent")] = f0["Ascent"]
        if "Descent" in f0:
            f1[Name("Descent")] = f0["Descent"]
        if "Leading" in f0:
            f1[Name("Leading")] = f0["Leading"]
        if "CapHeight" in f0:
            f1[Name("CapHeight")] = f0["CapHeight"]
        if "XHeight" in f0:
            f1[Name("XHeight")] = f0["XHeight"]
        if "StemV" in f0:
            f1[Name("StemV")] = f0["StemV"]
        if "StemH" in f0:
            f1[Name("StemH")] = f0["StemH"]
        if "AvgWidth" in f0:
            f1[Name("AvgWidth")] = f0["AvgWidth"]
        if "MaxWidth" in f0:
            f1[Name("MaxWidth")] = f0["MaxWidth"]
        if "MissingWidth" in f0:
            f1[Name("MissingWidth")] = f0["MissingWidth"]
        if "FontFile" in f0 and False:  # TODO
            f1[Name("FontFile")] = copy.deepcopy(f0["FontFile"])
        if "FontFile2" in f0 and False:  # TODO
            f1[Name("FontFile2")] = copy.deepcopy(f0["FontFile2"])
        if "FontFile3" in f0 and False:  # TODO
            f1[Name("FontFile3")] = copy.deepcopy(f0["FontFile3"])
        if "CharSet" in f0 and False:  # TODO
            f1[Name("CharSet")] = f0["CharSet"]
        # default
        for k, v in f0.items():
            if k not in f1:
                f1[k] = copy.deepcopy(v)
        return f1



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of LayoutElement acts just like Paragraph.
It also adds an outline in the document outline tree.
"""
import typing
from decimal import Decimal

from borb.pdf.canvas.color.color import Color, HexColor
from borb.pdf.canvas.font.font import Font
from borb.pdf.canvas.geometry.rectangle import Rectangle
from borb.pdf.canvas.layout.annotation.link_annotation import DestinationType
from borb.pdf.canvas.layout.hyphenation.hyphenation import Hyphenation
from borb.pdf.canvas.layout.layout_element import Alignment
from borb.pdf.canvas.layout.text.paragraph import Paragraph
from borb.pdf.document.document import Document
from borb.pdf.page.page import Page


class Heading(Paragraph):
    """
    This implementation of LayoutElement acts just like Paragraph.
    It also adds an outline in the document outline tree.
    """

    @staticmethod
    def _get_margin_for_outline_level(
        outline_level: int = 0, font_size: Decimal = Decimal(12)
    ) -> Decimal:
        return {
            1: Decimal(0.335),
            2: Decimal(0.553),
            3: Decimal(0.855),
            4: Decimal(1.333),
            5: Decimal(2.012),
            6: Decimal(3.477),
        }.get(outline_level + 1, Decimal(1)) * font_size

    def __init__(
        self,
        text: str,
        background_color: typing.Optional[Color] = None,
        border_bottom: bool = False,
        border_color: Color = HexColor("000000"),
        border_left: bool = False,
        border_radius_bottom_left: Decimal = Decimal(0),
        border_radius_bottom_right: Decimal = Decimal(0),
        border_radius_top_left: Decimal = Decimal(0),
        border_radius_top_right: Decimal = Decimal(0),
        border_right: bool = False,
        border_top: bool = False,
        border_width: Decimal = Decimal(1),
        fixed_leading: typing.Optional[Decimal] = None,
        font: typing.Union[Font, str] = "Helvetica",
        font_color: Color = HexColor("000000"),
        font_size: Decimal = Decimal(12),
        horizontal_alignment: Alignment = Alignment.LEFT,
        hyphenation: typing.Optional[Hyphenation] = None,
        margin_bottom: typing.Optional[Decimal] = None,
        margin_left: typing.Optional[Decimal] = None,
        margin_right: typing.Optional[Decimal] = None,
        margin_top: typing.Optional[Decimal] = None,
        multiplied_leading: typing.Optional[Decimal] = None,
        outline_level: int = 0,
        outline_text: typing.Optional[str] = None,
        padding_bottom: Decimal = Decimal(0),
        padding_left: Decimal = Decimal(0),
        padding_right: Decimal = Decimal(0),
        padding_top: Decimal = Decimal(0),
        respect_newlines_in_text: bool = False,
        respect_spaces_in_text: bool = False,
        text_alignment: Alignment = Alignment.LEFT,
        vertical_alignment: Alignment = Alignment.TOP,
    ):
        super().__init__(
            background_color=background_color,
            border_bottom=border_bottom,
            border_color=border_color,
            border_left=border_left,
            border_radius_bottom_left=border_radius_bottom_left,
            border_radius_bottom_right=border_radius_bottom_right,
            border_radius_top_left=border_radius_top_left,
            border_radius_top_right=border_radius_top_right,
            border_right=border_right,
            border_top=border_top,
            border_width=border_width,
            fixed_leading=fixed_leading,
            font=font,
            font_color=font_color,
            font_size=font_size,
            horizontal_alignment=horizontal_alignment,
            hyphenation=hyphenation,
            margin_bottom=margin_bottom
            if margin_bottom is not None
            else Heading._get_margin_for_outline_level(outline_level, font_size),
            margin_left=margin_left or Decimal(0),
            margin_right=margin_right or Decimal(0),
            margin_top=margin_top
            if margin_top is not None
            else Heading._get_margin_for_outline_level(outline_level, font_size),
            multiplied_leading=multiplied_leading,
            padding_bottom=padding_bottom,
            padding_left=padding_left,
            padding_right=padding_right,
            padding_top=padding_top,
            respect_newlines_in_text=respect_newlines_in_text,
            respect_spaces_in_text=respect_spaces_in_text,
            text=text,
            text_alignment=text_alignment,
            vertical_alignment=vertical_alignment,
        )
        self._outline_text = outline_text or text
        self._outline_level = outline_level
        self._has_added_outline = False

    def paint(self, page: Page, available_space: Rectangle):
        """
        This method paints this LayoutElement on the given Page, in the available space
        :param page:                the Page on which to paint this LayoutElement
        :param available_space:     the available space (as a Rectangle) on which to paint this LayoutElement
        :return:                    None
        """

        # fetch document
        p = page.get_root()  # type: ignore[attr-defined]
        assert isinstance(p, Document)

        # add outline to document
        if not self._has_added_outline:
            page_nr = page.get_page_info().get_page_number()
            assert page_nr is not None
            p.add_outline(
                text=self._outline_text,
                level=self._outline_level,
                destination_type=DestinationType.FIT,
                page_nr=int(page_nr),
            )

        # mark
        self._has_added_outline = True

        # call super
        super(Heading, self).paint(page, available_space)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
(PDF 1.2) Same as SCN but used for nonstroking operations.
"""
import typing
from decimal import Decimal
from typing import List

from borb.io.read.types import AnyPDFType
from borb.pdf.canvas.color.color import CMYKColor, GrayColor, RGBColor, Separation
from borb.pdf.canvas.operator.canvas_operator import CanvasOperator


class SetColorNonStroking(CanvasOperator):
    """
    (PDF 1.2) Same as SCN but used for nonstroking operations.
    """

    def __init__(self, canvas_stream_processor: "CanvasStreamProcessor"):  # type: ignore [name-defined]
        super().__init__("scn", 0)
        self._canvas = canvas_stream_processor.get_canvas()

    def get_number_of_operands(self) -> int:
        """
        This function returns the number of operands for the scn operator.
        The number of operands and their interpretation depends on the colour space.
        """
        non_stroke_color_space = self._canvas.graphics_state.non_stroke_color_space
        if non_stroke_color_space == "DeviceCMYK":
            return 4
        if non_stroke_color_space == "DeviceGray":
            return 1
        if non_stroke_color_space == "DeviceRGB":
            return 3
        # separation
        if (
            isinstance(non_stroke_color_space, List)
            and len(non_stroke_color_space) == 4
            and non_stroke_color_space[0] == "Separation"
        ):
            return 1
        return self._number_of_operands

    def invoke(
        self,
        canvas_stream_processor: "CanvasStreamProcessor",  # type: ignore [name-defined]
        operands: typing.List[AnyPDFType] = [],
        event_listeners: typing.List["EventListener"] = [],  # type: ignore [name-defined]
    ) -> None:
        """
        Invoke the scn operator
        """
        canvas = canvas_stream_processor.get_canvas()
        non_stroke_color_space = canvas.graphics_state.non_stroke_color_space
        if non_stroke_color_space == "DeviceCMYK":
            # fmt: off
            assert isinstance(operands[0], Decimal), "Operand 0 of scn must be a Decimal"
            assert isinstance(operands[1], Decimal), "Operand 1 of scn must be a Decimal"
            assert isinstance(operands[2], Decimal), "Operand 2 of scn must be a Decimal"
            assert isinstance(operands[3], Decimal), "Operand 3 of scn must be a Decimal"
            canvas.graphics_state.non_stroke_color = CMYKColor(
                operands[0],
                operands[1],
                operands[2],
                operands[3],
            )
            return
            # fmt: on

        if non_stroke_color_space == "DeviceGray":
            # fmt: off
            assert isinstance(operands[0], Decimal), "Operand 0 of scn must be a Decimal"
            canvas.graphics_state.non_stroke_color = GrayColor(operands[0])
            return
            # fmt: on

        if non_stroke_color_space == "DeviceRGB":
            # fmt: off
            assert isinstance(operands[0], Decimal), "Operand 0 of scn must be a Decimal"
            assert isinstance(operands[1], Decimal), "Operand 1 of scn must be a Decimal"
            assert isinstance(operands[2], Decimal), "Operand 2 of scn must be a Decimal"
            canvas.graphics_state.non_stroke_color = RGBColor(
                operands[0],
                operands[1],
                operands[2],
            )
            return
            # fmt: on

        # separation
        if (
            isinstance(non_stroke_color_space, List)
            and non_stroke_color_space[0] == "Separation"
        ):
            # fmt: off
            assert isinstance(operands[0], Decimal), "Operand 0 of scn must be a Decimal"
            canvas.graphics_state.non_stroke_color = Separation(canvas.graphics_state.non_stroke_color_space, [operands[0]])
            return
            # fmt: on



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
This implementation of WriteBaseTransformer is responsible for writing Dictionary objects
"""
import logging
import typing
from typing import Optional

from PIL.Image import Image  # type: ignore [import]

from borb.io.read.types import AnyPDFType, Dictionary, Element, List, Reference, Stream
from borb.io.write.transformer import Transformer, WriteTransformerState

logger = logging.getLogger(__name__)


class DictionaryTransformer(Transformer):
    """
    This implementation of WriteBaseTransformer is responsible for writing Dictionary objects
    """

    def can_be_transformed(self, any: AnyPDFType):
        """
        This function returns True if the object to be converted represents an Dictionary object
        """
        return isinstance(any, Dictionary)

    def transform(
        self,
        object_to_transform: AnyPDFType,
        context: Optional[WriteTransformerState] = None,
    ):
        """
        This method writes a Dictionary to a byte stream
        """
        # fmt: off
        assert isinstance(object_to_transform, Dictionary), "object_to_transform must be of type Dictionary"
        assert (context is not None), "context must be defined in order to write Dictionary objects."
        assert context.destination is not None, "context.destination must be defined in order to write Dictionary objects."
        # fmt: on

        # avoid resolving objects twice
        object_ref: typing.Optional[Reference] = object_to_transform.get_reference()  # type: ignore [attr-defined]
        if object_ref is not None and object_ref in context.resolved_references:
            assert object_ref is not None
            assert object_ref.object_number is not None
            logger.debug(
                "skip writing object %d %d R (already resolved)"
                % (object_ref.object_number, object_ref.generation_number or 0)
            )
            return

        # output value
        out_value = Dictionary()

        # objects to turn into reference
        queue: typing.List[AnyPDFType] = []
        for k, v in object_to_transform.items():
            if (
                isinstance(v, Dictionary)
                or isinstance(v, List)
                or isinstance(v, Stream)
                or isinstance(v, Image)
                or isinstance(v, Element)
            ) and not v.is_inline():  # type: ignore [union-attr]
                out_value[k] = self.get_reference(v, context)
                queue.append(v)
            else:
                out_value[k] = v

        # start object if needed
        started_object = False
        if object_ref is not None:
            assert object_ref.object_number is not None
            if object_ref.object_number is not None and object_ref.byte_offset is None:
                started_object = True
                self._start_object(object_to_transform, context)
            context.resolved_references.append(object_ref)

        # write dictionary at current location
        context.destination.write(bytes("<<", "latin1"))
        N = len(out_value.items())
        for i, (k, v) in enumerate(out_value.items()):
            self.get_root_transformer().transform(k, context)
            context.destination.write(bytes(" ", "latin1"))
            self.get_root_transformer().transform(v, context)
            if i != N - 1:
                context.destination.write(bytes(" ", "latin1"))

        # write newline if the object is not inline
        if object_to_transform.is_inline():  # type: ignore [attr-defined]
            context.destination.write(bytes(">>", "latin1"))
        else:
            context.destination.write(bytes(">>\n", "latin1"))

        # end object if needed
        if started_object:
            self._end_object(object_to_transform, context)

        for e in queue:
            self.get_root_transformer().transform(e, context)

        # return
        return out_value
